#include <stdio.h>
#include <stdint.h>

#include "e6809.h"
#include "trace.h"

#include "vecx.h" // for dat FLAGS

/* code assumptions:
 *  - it is assumed that an 'int' is at least 16 bits long.
 *  - a 16-bit register has valid bits only in the lower 16 bits and an
 *    8-bit register has valid bits only in the lower 8 bits. the upper
 *    may contain garbage!
 *  - all reading functions are assumed to return the requested data in
 *    the lower bits with the unused upper bits all set to zero.
 */

enum
{
	FLAG_E = 0x80,
	FLAG_F = 0x40,
	FLAG_H = 0x20,
	FLAG_I = 0x10,
	FLAG_N = 0x08,
	FLAG_Z = 0x04,
	FLAG_V = 0x02,
	FLAG_C = 0x01,

	IRQ_NORMAL = 0,
	IRQ_SYNC = 1,
	IRQ_CWAI = 2
};

//M6809 CPU;

/* user defined read and write functions */

//uint8_t(*e6809_read8) (uint16_t address);
//void(*e6809_write8) (uint16_t address, uint8_t data);

/* obtain a particular condition code. returns 0 or 1. */

static __inline uint8_t get_cc(M6809 *CPU, uint8_t flag)
{
	return (CPU->reg_cc / flag) & 1;
}

/* set a particular condition code to either 0 or 1.
 * value parameter must be either 0 or 1.
 */

static __inline void set_cc(M6809 *CPU, uint8_t flag, uint8_t value)
{
	CPU->reg_cc &= ~flag;
	CPU->reg_cc |= value * flag;
}

/* test carry */

static __inline uint8_t test_c(uint16_t i0, uint16_t i1,
	uint16_t r, uint16_t sub)
{
	uint16_t flag;

	flag = (i0 | i1) & ~r; /* one of the inputs is 1 and output is 0 */
	flag |= (i0 & i1);      /* both inputs are 1 */
	flag = (flag >> 7) & 1;
	flag ^= sub; /* on a sub, carry is opposite the carry of an add */

	return (uint8_t)flag;
}

/* test negative */

static __inline uint8_t test_n(uint16_t r)
{
	return (r >> 7) & 1;
}

/* test for zero in lower 8 bits */

static __inline uint8_t test_z8(uint16_t r)
{
	uint16_t flag;

	flag = ~r;
	flag = (flag >> 4) & (flag & 0xf);
	flag = (flag >> 2) & (flag & 0x3);
	flag = (flag >> 1) & (flag & 0x1);

	return (uint8_t)flag;
}

/* test for zero in lower 16 bits */

static __inline uint8_t test_z16(uint16_t r)
{
	uint16_t flag;

	flag = ~r;
	flag = (flag >> 8) & (flag & 0xff);
	flag = (flag >> 4) & (flag & 0xf);
	flag = (flag >> 2) & (flag & 0x3);
	flag = (flag >> 1) & (flag & 0x1);

	return (uint8_t)flag;
}

/* overflow is set whenever the sign bits of the inputs are the same
 * but the sign bit of the result is not same as the sign bits of the
 * inputs.
 */

static __inline uint8_t test_v(uint16_t i0, uint16_t i1, uint16_t r)
{
	uint16_t flag;

	flag = ~(i0 ^ i1); /* input sign bits are the same */
	flag &= (i0 ^ r);  /* input sign and output sign not same */
	flag = (flag >> 7) & 1;

	return (uint8_t)flag;
}

static __inline uint16_t get_reg_d(M6809 *CPU)
{
	return (CPU->reg_a << 8) | (CPU->reg_b & 0xff);
}

static __inline void set_reg_d(M6809 *CPU, uint16_t value)
{
	CPU->reg_a = (uint8_t)(value >> 8);
	CPU->reg_b = (uint8_t)value;
}

/* read a byte ... the returned value has the lower 8-bits set to the byte
 * while the upper bits are all zero.
 */

static __inline uint8_t read8(M6809 *CPU, uint16_t address)
{
	return (*CPU->read8) (CPU->userdata, address);
}

/* write a byte ... only the lower 8-bits of the uint16_t data
 * is written. the upper bits are ignored.
 */

static __inline void write8(M6809 *CPU, uint16_t address, uint8_t data)
{
    (*CPU->write8)(CPU->userdata, address, data);
}

static __inline uint16_t read16(M6809 *CPU, uint16_t address)
{
	return (read8(CPU, address) << 8) | read8(CPU, address + 1);
}

static __inline void write16(M6809 *CPU, uint16_t address, uint16_t data)
{
	write8(CPU, address, (uint8_t)(data >> 8));
	write8(CPU, address + 1, (uint8_t)data);
}

static __inline void push8(M6809 *CPU, uint16_t *sp, uint8_t data)
{
	(*sp)--;
	write8(CPU, *sp, data);
}

static __inline uint8_t pull8(M6809 *CPU, uint16_t *sp)
{
	uint8_t data;

	data = read8(CPU, *sp);
	(*sp)++;

	return data;
}

static __inline void push16(M6809 *CPU, uint16_t *sp, uint16_t data)
{
	push8(CPU, sp, (uint8_t)data);
	push8(CPU, sp, (uint8_t)(data >> 8));
}

static __inline uint16_t pull16(M6809 *CPU, uint16_t *sp)
{
	uint16_t tmp = pull8(CPU, sp);
	return (tmp << 8) | pull8(CPU, sp);
}

/* read a byte from the address pointed to by the pc */

static __inline uint8_t pc_read8(M6809 *CPU)
{
        CPU->F[CPU->reg_pc] |= F_OPERAND;
        // If same or unassigned, note PC.  If assigned and different, note that impossible to infer SETDP
        CPU->DP[CPU->reg_pc] = (CPU->DP[CPU->reg_pc] == CPU->reg_dp || CPU->DP[CPU->reg_pc] == 0x100) ? CPU->reg_dp : 0x200;
	return read8(CPU, CPU->reg_pc++);
}

/* read a word from the address pointed to by the pc */

static __inline uint16_t pc_read16(M6809 *CPU)
{
	uint16_t data;

	data = read16(CPU, CPU->reg_pc);
	CPU->reg_pc += 2;

	return data;
}

/* sign extend an 8-bit quantity into a 16-bit quantity */

static __inline uint16_t sign_extend(uint16_t data)
{
	return (~(data & 0x80) + 1) | (data & 0xff);
}

/* direct addressing, upper byte of the address comes from
 * the direct page register, and the lower byte comes from the
 * instruction itself.
 */

static __inline uint16_t ea_direct(M6809 *CPU)
{
	return (CPU->reg_dp << 8) | pc_read8(CPU);
}

/* extended addressing, address is obtained from 2 bytes following
 * the instruction.
 */

static __inline uint16_t ea_extended(M6809 *CPU)
{
	return pc_read16(CPU);
}

/* indexed addressing */

static __inline uint16_t ea_indexed(M6809 *CPU, uint16_t *cycles)
{
	uint16_t ea = 0;

	/* post byte */

	uint16_t op = pc_read8(CPU);

	uint16_t r = (op >> 5) & 3;

	switch (op)
	{
	case 0x00: case 0x01: case 0x02: case 0x03:
	case 0x04: case 0x05: case 0x06: case 0x07:
	case 0x08: case 0x09: case 0x0a: case 0x0b:
	case 0x0c: case 0x0d: case 0x0e: case 0x0f:
	case 0x20: case 0x21: case 0x22: case 0x23:
	case 0x24: case 0x25: case 0x26: case 0x27:
	case 0x28: case 0x29: case 0x2a: case 0x2b:
	case 0x2c: case 0x2d: case 0x2e: case 0x2f:
	case 0x40: case 0x41: case 0x42: case 0x43:
	case 0x44: case 0x45: case 0x46: case 0x47:
	case 0x48: case 0x49: case 0x4a: case 0x4b:
	case 0x4c: case 0x4d: case 0x4e: case 0x4f:
	case 0x60: case 0x61: case 0x62: case 0x63:
	case 0x64: case 0x65: case 0x66: case 0x67:
	case 0x68: case 0x69: case 0x6a: case 0x6b:
	case 0x6c: case 0x6d: case 0x6e: case 0x6f:
		/* R, +[0, 15] */

		ea = *CPU->rptr_xyus[r] + (op & 0xf);
		(*cycles)++;
		break;
	case 0x10: case 0x11: case 0x12: case 0x13:
	case 0x14: case 0x15: case 0x16: case 0x17:
	case 0x18: case 0x19: case 0x1a: case 0x1b:
	case 0x1c: case 0x1d: case 0x1e: case 0x1f:
	case 0x30: case 0x31: case 0x32: case 0x33:
	case 0x34: case 0x35: case 0x36: case 0x37:
	case 0x38: case 0x39: case 0x3a: case 0x3b:
	case 0x3c: case 0x3d: case 0x3e: case 0x3f:
	case 0x50: case 0x51: case 0x52: case 0x53:
	case 0x54: case 0x55: case 0x56: case 0x57:
	case 0x58: case 0x59: case 0x5a: case 0x5b:
	case 0x5c: case 0x5d: case 0x5e: case 0x5f:
	case 0x70: case 0x71: case 0x72: case 0x73:
	case 0x74: case 0x75: case 0x76: case 0x77:
	case 0x78: case 0x79: case 0x7a: case 0x7b:
	case 0x7c: case 0x7d: case 0x7e: case 0x7f:
		/* R, +[-16, -1] */

		ea = *CPU->rptr_xyus[r] + (op & 0xf) - 0x10;
		(*cycles)++;
		break;
	case 0x80: case 0x81:
	case 0xa0: case 0xa1:
	case 0xc0: case 0xc1:
	case 0xe0: case 0xe1:
		/* ,R+ / ,R++ */

		ea = *CPU->rptr_xyus[r];
		*CPU->rptr_xyus[r] += 1 + (op & 1);
		*cycles += 2 + (op & 1);
		break;
	case 0x90: case 0x91:
	case 0xb0: case 0xb1:
	case 0xd0: case 0xd1:
	case 0xf0: case 0xf1:
		/* [,R+] ??? / [,R++] */

		ea = read16(CPU, *CPU->rptr_xyus[r]);
		*CPU->rptr_xyus[r] += 1 + (op & 1);
		*cycles += 5 + (op & 1);
		break;
	case 0x82: case 0x83:
	case 0xa2: case 0xa3:
	case 0xc2: case 0xc3:
	case 0xe2: case 0xe3:

		/* ,-R / ,--R */

		*CPU->rptr_xyus[r] -= 1 + (op & 1);
		ea = *CPU->rptr_xyus[r];
		*cycles += 2 + (op & 1);
		break;
	case 0x92: case 0x93:
	case 0xb2: case 0xb3:
	case 0xd2: case 0xd3:
	case 0xf2: case 0xf3:
		/* [,-R] ??? / [,--R] */

		*CPU->rptr_xyus[r] -= 1 + (op & 1);
		ea = read16(CPU, *CPU->rptr_xyus[r]);
		*cycles += 5 + (op & 1);
		break;
	case 0x84: case 0xa4:
	case 0xc4: case 0xe4:
		/* ,R */

		ea = *CPU->rptr_xyus[r];
		break;
	case 0x94: case 0xb4:
	case 0xd4: case 0xf4:
		/* [,R] */

		ea = read16(CPU, *CPU->rptr_xyus[r]);
		*cycles += 3;
		break;
	case 0x85: case 0xa5:
	case 0xc5: case 0xe5:
		/* B,R */

		ea = *CPU->rptr_xyus[r] + sign_extend(CPU->reg_b);
		*cycles += 1;
		break;
	case 0x95: case 0xb5:
	case 0xd5: case 0xf5:
		/* [B,R] */

		ea = read16(CPU, *CPU->rptr_xyus[r] + sign_extend(CPU->reg_b));
		*cycles += 4;
		break;
	case 0x86: case 0xa6:
	case 0xc6: case 0xe6:
		/* A,R */

		ea = *CPU->rptr_xyus[r] + sign_extend(CPU->reg_a);
		*cycles += 1;
		break;
	case 0x96: case 0xb6:
	case 0xd6: case 0xf6:
		/* [A,R] */

		ea = read16(CPU, *CPU->rptr_xyus[r] + sign_extend(CPU->reg_a));
		*cycles += 4;
		break;
	case 0x88: case 0xa8:
	case 0xc8: case 0xe8:
		/* byte,R */

		ea = *CPU->rptr_xyus[r] + sign_extend(pc_read8(CPU));
		*cycles += 1;
		break;
	case 0x98: case 0xb8:
	case 0xd8: case 0xf8:
		/* [byte,R] */

		ea = read16(CPU, *CPU->rptr_xyus[r] + sign_extend(pc_read8(CPU)));
		*cycles += 4;
		break;
	case 0x89: case 0xa9:
	case 0xc9: case 0xe9:
		/* word,R */

		ea = *CPU->rptr_xyus[r] + pc_read16(CPU);
		*cycles += 4;
		break;
	case 0x99: case 0xb9:
	case 0xd9: case 0xf9:
		/* [word,R] */

		ea = read16(CPU, *CPU->rptr_xyus[r] + pc_read16(CPU));
		*cycles += 7;
		break;
	case 0x8b: case 0xab:
	case 0xcb: case 0xeb:
		/* D,R */

		ea = *CPU->rptr_xyus[r] + get_reg_d(CPU);
		*cycles += 4;
		break;
	case 0x9b: case 0xbb:
	case 0xdb: case 0xfb:
		/* [D,R] */

		ea = read16(CPU, *CPU->rptr_xyus[r] + get_reg_d(CPU));
		*cycles += 7;
		break;
	case 0x8c: case 0xac:
	case 0xcc: case 0xec:
		/* byte, PC */

		r = sign_extend(pc_read8(CPU));
		ea = CPU->reg_pc + r;
		*cycles += 1;
		break;
	case 0x9c: case 0xbc:
	case 0xdc: case 0xfc:
		/* [byte, PC] */

		r = sign_extend(pc_read8(CPU));
		ea = read16(CPU, CPU->reg_pc + r);
		*cycles += 4;
		break;
	case 0x8d: case 0xad:
	case 0xcd: case 0xed:
		/* word, PC */

		r = pc_read16(CPU);
		ea = CPU->reg_pc + r;
		*cycles += 5;
		break;
	case 0x9d: case 0xbd:
	case 0xdd: case 0xfd:
		/* [word, PC] */

		r = pc_read16(CPU);
		ea = read16(CPU, CPU->reg_pc + r);
		*cycles += 8;
		break;
	case 0x9f:
		/* [address] */

		ea = read16(CPU, pc_read16(CPU));
		*cycles += 5;
		break;
	default:
		printf("undefined post-byte\n");
		break;
	}

	return ea;
}

/* instruction: neg
 * essentially (0 - data).
 */

static __inline uint16_t inst_neg(M6809 *CPU, uint16_t data)
{
	uint16_t i0 = 0;
	uint16_t i1 = ~data;
	uint16_t r = i0 + i1 + 1;

	set_cc(CPU, FLAG_H, test_c(i0 << 4, i1 << 4, r << 4, 0));
	set_cc(CPU, FLAG_N, test_n(r));
	set_cc(CPU, FLAG_Z, test_z8(r));
	set_cc(CPU, FLAG_V, test_v(i0, i1, r));
	set_cc(CPU, FLAG_C, test_c(i0, i1, r, 1));

	return r;
}

/* instruction: com */

static __inline uint16_t inst_com(M6809 *CPU, uint16_t data)
{
	uint16_t r = ~data;

	set_cc(CPU, FLAG_N, test_n(r));
	set_cc(CPU, FLAG_Z, test_z8(r));
	set_cc(CPU, FLAG_V, 0);
	set_cc(CPU, FLAG_C, 1);

	return r;
}

/* instruction: lsr
 * cannot be faked as an add or substract.
 */

static __inline uint16_t inst_lsr(M6809 *CPU, uint16_t data)
{
	uint16_t r = (data >> 1) & 0x7f;

	set_cc(CPU, FLAG_N, 0);
	set_cc(CPU, FLAG_Z, test_z8(r));
	set_cc(CPU, FLAG_C, data & 1);

	return r;
}

/* instruction: ror
 * cannot be faked as an add or substract.
 */

static __inline uint16_t inst_ror(M6809 *CPU, uint16_t data)
{
	uint16_t c = get_cc(CPU, FLAG_C);
	uint16_t r = ((data >> 1) & 0x7f) | (c << 7);

	set_cc(CPU, FLAG_N, test_n(r));
	set_cc(CPU, FLAG_Z, test_z8(r));
	set_cc(CPU, FLAG_C, data & 1);

	return r;
}

/* instruction: asr
 * cannot be faked as an add or substract.
 */

static __inline uint16_t inst_asr(M6809 *CPU, uint16_t data)
{
	uint16_t r = ((data >> 1) & 0x7f) | (data & 0x80);

	set_cc(CPU, FLAG_N, test_n(r));
	set_cc(CPU, FLAG_Z, test_z8(r));
	set_cc(CPU, FLAG_C, data & 1);

	return r;
}

/* instruction: asl
 * essentially (data + data). simple addition.
 */

static __inline uint16_t inst_asl(M6809 *CPU, uint16_t data)
{
	uint16_t i0 = data;
	uint16_t i1 = data;
	uint16_t r = i0 + i1;

	set_cc(CPU, FLAG_H, test_c(i0 << 4, i1 << 4, r << 4, 0));
	set_cc(CPU, FLAG_N, test_n(r));
	set_cc(CPU, FLAG_Z, test_z8(r));
	set_cc(CPU, FLAG_V, test_v(i0, i1, r));
	set_cc(CPU, FLAG_C, test_c(i0, i1, r, 0));

	return r;
}

/* instruction: rol
 * essentially (data + data + carry). addition with carry.
 */

static __inline uint16_t inst_rol(M6809 *CPU, uint16_t data)
{
	uint16_t i0 = data;
	uint16_t i1 = data;
	uint16_t c = get_cc(CPU, FLAG_C);
	uint16_t r = i0 + i1 + c;

	set_cc(CPU, FLAG_N, test_n(r));
	set_cc(CPU, FLAG_Z, test_z8(r));
	set_cc(CPU, FLAG_V, test_v(i0, i1, r));
	set_cc(CPU, FLAG_C, test_c(i0, i1, r, 0));

	return r;
}

/* instruction: dec
 * essentially (data - 1).
 */

static __inline uint16_t inst_dec(M6809 *CPU, uint16_t data)
{
	uint16_t i0 = data;
	uint16_t i1 = 0xff;
	uint16_t r = i0 + i1;

	set_cc(CPU, FLAG_N, test_n(r));
	set_cc(CPU, FLAG_Z, test_z8(r));
	set_cc(CPU, FLAG_V, test_v(i0, i1, r));

	return r;
}

/* instruction: inc
 * essentially (data + 1).
 */

static __inline uint16_t inst_inc(M6809 *CPU, uint16_t data)
{
	uint16_t i0 = data;
	uint16_t i1 = 1;
	uint16_t r = i0 + i1;

	set_cc(CPU, FLAG_N, test_n(r));
	set_cc(CPU, FLAG_Z, test_z8(r));
	set_cc(CPU, FLAG_V, test_v(i0, i1, r));

	return r;
}

/* instruction: tst */

static __inline void inst_tst8(M6809 *CPU, uint8_t data)
{
	set_cc(CPU, FLAG_N, test_n(data));
	set_cc(CPU, FLAG_Z, test_z8(data));
	set_cc(CPU, FLAG_V, 0);
}

static __inline void inst_tst16(M6809 *CPU, uint16_t data)
{
	set_cc(CPU, FLAG_N, test_n(data >> 8));
	set_cc(CPU, FLAG_Z, test_z16(data));
	set_cc(CPU, FLAG_V, 0);
}

/* instruction: clr */

static __inline void inst_clr(M6809 *CPU)
{
	set_cc(CPU, FLAG_N, 0);
	set_cc(CPU, FLAG_Z, 1);
	set_cc(CPU, FLAG_V, 0);
	set_cc(CPU, FLAG_C, 0);
}

/* instruction: suba/subb */

static __inline uint8_t inst_sub8(M6809 *CPU, uint8_t data0, uint8_t data1)
{
	uint8_t i0 = data0;
	uint8_t i1 = ~data1;
	uint8_t r = i0 + i1 + 1;

	set_cc(CPU, FLAG_H, test_c(i0 << 4, i1 << 4, r << 4, 0));
	set_cc(CPU, FLAG_N, test_n(r));
	set_cc(CPU, FLAG_Z, test_z8(r));
	set_cc(CPU, FLAG_V, test_v(i0, i1, r));
	set_cc(CPU, FLAG_C, test_c(i0, i1, r, 1));

	return r;
}

/* instruction: sbca/sbcb/cmpa/cmpb.
 * only 8-bit version, 16-bit version not needed.
 */

static __inline uint8_t inst_sbc(M6809 *CPU, uint8_t data0, uint8_t data1)
{
	uint8_t i0 = data0;
	uint8_t i1 = ~data1;
	uint8_t c = 1 - get_cc(CPU, FLAG_C);
	uint8_t r = i0 + i1 + c;

	set_cc(CPU, FLAG_H, test_c(i0 << 4, i1 << 4, r << 4, 0));
	set_cc(CPU, FLAG_N, test_n(r));
	set_cc(CPU, FLAG_Z, test_z8(r));
	set_cc(CPU, FLAG_V, test_v(i0, i1, r));
	set_cc(CPU, FLAG_C, test_c(i0, i1, r, 1));

	return r;
}

/* instruction: anda/andb/bita/bitb.
 * only 8-bit version, 16-bit version not needed.
 */

static __inline uint8_t inst_and(M6809 *CPU, uint8_t data0, uint8_t data1)
{
	uint8_t r = data0 & data1;

	inst_tst8(CPU, r);

	return r;
}

/* instruction: eora/eorb.
 * only 8-bit version, 16-bit version not needed.
 */

static __inline uint8_t inst_eor(M6809 *CPU, uint8_t data0, uint8_t data1)
{
	uint8_t r = data0 ^ data1;

	inst_tst8(CPU, r);

	return r;
}

/* instruction: adca/adcb
 * only 8-bit version, 16-bit version not needed.
 */

static __inline uint8_t inst_adc(M6809 *CPU, uint8_t data0, uint8_t data1)
{
	uint8_t i0 = data0;
	uint8_t i1 = data1;
	uint8_t c = get_cc(CPU, FLAG_C);
	uint8_t r = i0 + i1 + c;

	set_cc(CPU, FLAG_H, test_c(i0 << 4, i1 << 4, r << 4, 0));
	set_cc(CPU, FLAG_N, test_n(r));
	set_cc(CPU, FLAG_Z, test_z8(r));
	set_cc(CPU, FLAG_V, test_v(i0, i1, r));
	set_cc(CPU, FLAG_C, test_c(i0, i1, r, 0));

	return r;
}

/* instruction: ora/orb.
 * only 8-bit version, 16-bit version not needed.
 */

static __inline uint8_t inst_or(M6809 *CPU, uint8_t data0, uint8_t data1)
{
	uint8_t r = data0 | data1;

	inst_tst8(CPU, r);

	return r;
}

/* instruction: adda/addb */

static __inline uint8_t inst_add8(M6809 *CPU, uint8_t data0, uint8_t data1)
{
	uint8_t i0 = data0;
	uint8_t i1 = data1;
	uint8_t r = i0 + i1;

	set_cc(CPU, FLAG_H, test_c(i0 << 4, i1 << 4, r << 4, 0));
	set_cc(CPU, FLAG_N, test_n(r));
	set_cc(CPU, FLAG_Z, test_z8(r));
	set_cc(CPU, FLAG_V, test_v(i0, i1, r));
	set_cc(CPU, FLAG_C, test_c(i0, i1, r, 0));

	return (uint8_t)r;
}

/* instruction: addd */

static __inline uint16_t inst_add16(M6809 *CPU, uint16_t data0, uint16_t data1)
{
	uint16_t i0 = data0;
	uint16_t i1 = data1;
	uint16_t r = i0 + i1;

	set_cc(CPU, FLAG_N, test_n(r >> 8));
	set_cc(CPU, FLAG_Z, test_z16(r));
	set_cc(CPU, FLAG_V, test_v(i0 >> 8, i1 >> 8, r >> 8));
	set_cc(CPU, FLAG_C, test_c(i0 >> 8, i1 >> 8, r >> 8, 0));

	return r;
}

/* instruction: subd */

static __inline uint16_t inst_sub16(M6809 *CPU, uint16_t data0, uint16_t data1)
{
	uint16_t i0 = data0;
	uint16_t i1 = ~data1;
	uint16_t r = i0 + i1 + 1;

	set_cc(CPU, FLAG_N, test_n(r >> 8));
	set_cc(CPU, FLAG_Z, test_z16(r));
	set_cc(CPU, FLAG_V, test_v(i0 >> 8, i1 >> 8, r >> 8));
	set_cc(CPU, FLAG_C, test_c(i0 >> 8, i1 >> 8, r >> 8, 1));

	return r;
}

/* instruction: 8-bit offset branch */

static __inline void inst_bra8(M6809 *CPU, uint16_t test, uint16_t op, uint16_t *cycles)
{
	uint16_t offset = pc_read8(CPU);

	/* trying to avoid an if statement */

	uint16_t mask = (test ^ (op & 1)) - 1; /* 0xffff when taken, 0 when not taken */
	CPU->reg_pc += sign_extend(offset) & mask;

	*cycles += 3;
}

/* instruction: 16-bit offset branch */

static __inline void inst_bra16(M6809 *CPU, uint16_t test, uint16_t op, uint16_t *cycles)
{
	uint16_t offset = pc_read16(CPU);

	/* trying to avoid an if statement */

	uint16_t mask = (test ^ (op & 1)) - 1; /* 0xffff when taken, 0 when not taken */
	CPU->reg_pc += offset & mask;

	*cycles += 5 - mask;
}

/* instruction: pshs/pshu */

static __inline void inst_psh(M6809 *CPU, uint16_t op, uint16_t *sp,
	uint16_t data, uint16_t *cycles)
{
	if (op & 0x80)
	{
		push16(CPU, sp, CPU->reg_pc);
		*cycles += 2;
	}

	if (op & 0x40)
	{
		/* either s or u */
		push16(CPU, sp, data);
		*cycles += 2;
	}

	if (op & 0x20)
	{
		push16(CPU, sp, CPU->reg_y);
		*cycles += 2;
	}

	if (op & 0x10)
	{
		push16(CPU, sp, CPU->reg_x);
		*cycles += 2;
	}

	if (op & 0x08)
	{
		push8(CPU, sp, CPU->reg_dp);
		*cycles += 1;
	}

	if (op & 0x04)
	{
		push8(CPU, sp, CPU->reg_b);
		*cycles += 1;
	}

	if (op & 0x02)
	{
		push8(CPU, sp, CPU->reg_a);
		*cycles += 1;
	}

	if (op & 0x01)
	{
		push8(CPU, sp, CPU->reg_cc);
		*cycles += 1;
	}
}

/* instruction: puls/pulu */

static __inline void inst_pul(M6809 *CPU, uint16_t op, uint16_t *sp, uint16_t *osp,
	uint16_t *cycles)
{
	if (op & 0x01)
	{
		CPU->reg_cc = pull8(CPU, sp);
		*cycles += 1;
	}

	if (op & 0x02)
	{
		CPU->reg_a = pull8(CPU, sp);
		*cycles += 1;
	}

	if (op & 0x04)
	{
		CPU->reg_b = pull8(CPU, sp);
		*cycles += 1;
	}

	if (op & 0x08)
	{
		CPU->reg_dp = pull8(CPU, sp);
		*cycles += 1;
	}

	if (op & 0x10)
	{
		CPU->reg_x = pull16(CPU, sp);
		*cycles += 2;
	}

	if (op & 0x20)
	{
		CPU->reg_y = pull16(CPU, sp);
		*cycles += 2;
	}

	if (op & 0x40)
	{
		/* either s or u */
		*osp = pull16(CPU, sp);
		*cycles += 2;
	}

	if (op & 0x80)
	{
		CPU->reg_pc = pull16(CPU, sp);
		*cycles += 2;
	}
}

static __inline uint16_t exgtfr_read(M6809 *CPU, uint16_t reg)
{
	uint16_t data;

	switch (reg)
	{
	case 0x0:
		data = get_reg_d(CPU);
		break;
	case 0x1:
		data = CPU->reg_x;
		break;
	case 0x2:
		data = CPU->reg_y;
		break;
	case 0x3:
		data = CPU->reg_u;
		break;
	case 0x4:
		data = CPU->reg_s;
		break;
	case 0x5:
		data = CPU->reg_pc;
		break;
	case 0x8:
		data = 0xff00 | CPU->reg_a;
		break;
	case 0x9:
		data = 0xff00 | CPU->reg_b;
		break;
	case 0xa:
		data = 0xff00 | CPU->reg_cc;
		break;
	case 0xb:
		data = 0xff00 | CPU->reg_dp;
		break;
	default:
		data = 0xffff;
		printf("illegal exgtfr reg %.1x\n", reg);
		break;
	}

	return data;
}

static __inline void exgtfr_write(M6809 *CPU, uint16_t reg, uint16_t data)
{
	switch (reg)
	{
	case 0x0:
		set_reg_d(CPU, data);
		break;
	case 0x1:
		CPU->reg_x = data;
		break;
	case 0x2:
		CPU->reg_y = data;
		break;
	case 0x3:
		CPU->reg_u = data;
		break;
	case 0x4:
		CPU->reg_s = data;
		break;
	case 0x5:
		CPU->reg_pc = data;
		break;
	case 0x8:
		CPU->reg_a = (uint8_t)data;
		break;
	case 0x9:
		CPU->reg_b = (uint8_t)data;
		break;
	case 0xa:
		CPU->reg_cc = (uint8_t)data;
		break;
	case 0xb:
		CPU->reg_dp = (uint8_t)data;
		break;
	default:
		printf("illegal exgtfr reg %.1x\n", reg);
		break;
	}
}

/* instruction: exg */

static __inline void inst_exg(M6809 *CPU)
{
	uint16_t op = pc_read8(CPU);

	uint16_t tmp = exgtfr_read(CPU, op & 0xf);
	exgtfr_write(CPU, op & 0xf, exgtfr_read(CPU, op >> 4));
	exgtfr_write(CPU, op >> 4, tmp);
}

/* instruction: tfr */

static __inline void inst_tfr(M6809 *CPU)
{
	uint16_t op = pc_read8(CPU);

	exgtfr_write(CPU, op & 0xf, exgtfr_read(CPU, op >> 4));
}

/* reset the 6809 */

void e6809_reset(M6809 *CPU)
{
	CPU->reg_x = 0;
	CPU->reg_y = 0;
	CPU->reg_u = 0;
	CPU->reg_s = 0;

	CPU->reg_a = 0;
	CPU->reg_b = 0;

	CPU->reg_dp = 0;

	CPU->reg_cc = FLAG_I | FLAG_F;
	CPU->irq_status = IRQ_NORMAL;

	CPU->rptr_xyus[0] = &CPU->reg_x;
	CPU->rptr_xyus[1] = &CPU->reg_y;
	CPU->rptr_xyus[2] = &CPU->reg_u;
	CPU->rptr_xyus[3] = &CPU->reg_s;

	CPU->reg_pc = read16(CPU, 0xfffe);
}

/* execute a single instruction or handle interrupts and return */

uint16_t e6809_sstep(M6809 *CPU, uint16_t irq_i, uint16_t irq_f)
{
	uint16_t cycles = 0;
        trace_instruction(CPU, irq_i, irq_f);
        
	if (irq_f)
	{
		if (get_cc(CPU, FLAG_F) == 0)
		{
			if (CPU->irq_status != IRQ_CWAI)
			{
				set_cc(CPU, FLAG_E, 0);
				inst_psh(CPU, 0x81, &CPU->reg_s, CPU->reg_u, &cycles);
			}

			set_cc(CPU, FLAG_I, 1);
			set_cc(CPU, FLAG_F, 1);

			CPU->reg_pc = read16(CPU, 0xfff6);
			CPU->irq_status = IRQ_NORMAL;
			cycles += 7;
		}
		else
		{
			if (CPU->irq_status == IRQ_SYNC)
			{
				CPU->irq_status = IRQ_NORMAL;
			}
		}
	}

	if (irq_i)
	{
		if (get_cc(CPU, FLAG_I) == 0)
		{
			if (CPU->irq_status != IRQ_CWAI)
			{
				set_cc(CPU, FLAG_E, 1);
				inst_psh(CPU, 0xff, &CPU->reg_s, CPU->reg_u, &cycles);
			}

			set_cc(CPU, FLAG_I, 1);

			CPU->reg_pc = read16(CPU, 0xfff8);
			CPU->irq_status = IRQ_NORMAL;
			cycles += 7;
		}
		else
		{
			if (CPU->irq_status == IRQ_SYNC)
			{
				CPU->irq_status = IRQ_NORMAL;
			}
		}
	}

	if (CPU->irq_status != IRQ_NORMAL)
	{
		return cycles + 1;
	}

	uint16_t op = pc_read8(CPU);
	uint16_t ea, i0, i1, r;
        CPU->F[CPU->reg_pc-1] |= F_OPCODE; // pc_read8 has marked it as an operand and incremented PC. Undo.

	switch (op)
	{
		/* page 0 instructions */

		/* neg, nega, negb */
	case 0x00:
		ea = ea_direct(CPU);
		r = inst_neg(CPU, read8(CPU, ea));
		write8(CPU, ea, (uint8_t)r);
		cycles += 6;
		break;
	case 0x40:
		CPU->reg_a = (uint8_t)inst_neg(CPU, CPU->reg_a);
		cycles += 2;
		break;
	case 0x50:
		CPU->reg_b = (uint8_t)inst_neg(CPU, CPU->reg_b);
		cycles += 2;
		break;
	case 0x60:
		ea = ea_indexed(CPU, &cycles);
		r = inst_neg(CPU, read8(CPU, ea));
		write8(CPU, ea, (uint8_t)r);
		cycles += 6;
		break;
	case 0x70:
		ea = ea_extended(CPU);
		r = inst_neg(CPU, read8(CPU, ea));
		write8(CPU, ea, (uint8_t)r);
		cycles += 7;
		break;
		/* com, coma, comb */
	case 0x03:
		ea = ea_direct(CPU);
		r = inst_com(CPU, read8(CPU, ea));
		write8(CPU, ea, (uint8_t)r);
		cycles += 6;
		break;
	case 0x43:
		CPU->reg_a = (uint8_t)inst_com(CPU, CPU->reg_a);
		cycles += 2;
		break;
	case 0x53:
		CPU->reg_b = (uint8_t)inst_com(CPU, CPU->reg_b);
		cycles += 2;
		break;
	case 0x63:
		ea = ea_indexed(CPU, &cycles);
		r = inst_com(CPU, read8(CPU, ea));
		write8(CPU, ea, (uint8_t)r);
		cycles += 6;
		break;
	case 0x73:
		ea = ea_extended(CPU);
		r = inst_com(CPU, read8(CPU, ea));
		write8(CPU, ea, (uint8_t)r);
		cycles += 7;
		break;
		/* lsr, lsra, lsrb */
	case 0x04:
		ea = ea_direct(CPU);
		r = inst_lsr(CPU, read8(CPU, ea));
		write8(CPU, ea, (uint8_t)r);
		cycles += 6;
		break;
	case 0x44:
		CPU->reg_a = (uint8_t)inst_lsr(CPU, CPU->reg_a);
		cycles += 2;
		break;
	case 0x54:
		CPU->reg_b = (uint8_t)inst_lsr(CPU, CPU->reg_b);
		cycles += 2;
		break;
	case 0x64:
		ea = ea_indexed(CPU, &cycles);
		r = inst_lsr(CPU, read8(CPU, ea));
		write8(CPU, ea, (uint8_t)r);
		cycles += 6;
		break;
	case 0x74:
		ea = ea_extended(CPU);
		r = inst_lsr(CPU, read8(CPU, ea));
		write8(CPU, ea, (uint8_t)r);
		cycles += 7;
		break;
		/* ror, rora, rorb */
	case 0x06:
		ea = ea_direct(CPU);
		r = inst_ror(CPU, read8(CPU, ea));
		write8(CPU, ea, (uint8_t)r);
		cycles += 6;
		break;
	case 0x46:
		CPU->reg_a = (uint8_t)inst_ror(CPU, CPU->reg_a);
		cycles += 2;
		break;
	case 0x56:
		CPU->reg_b = (uint8_t)inst_ror(CPU, CPU->reg_b);
		cycles += 2;
		break;
	case 0x66:
		ea = ea_indexed(CPU, &cycles);
		r = inst_ror(CPU, read8(CPU, ea));
		write8(CPU, ea, (uint8_t)r);
		cycles += 6;
		break;
	case 0x76:
		ea = ea_extended(CPU);
		r = inst_ror(CPU, read8(CPU, ea));
		write8(CPU, ea, (uint8_t)r);
		cycles += 7;
		break;
		/* asr, asra, asrb */
	case 0x07:
		ea = ea_direct(CPU);
		r = inst_asr(CPU, read8(CPU, ea));
		write8(CPU, ea, (uint8_t)r);
		cycles += 6;
		break;
	case 0x47:
		CPU->reg_a = (uint8_t)inst_asr(CPU, CPU->reg_a);
		cycles += 2;
		break;
	case 0x57:
		CPU->reg_b = (uint8_t)inst_asr(CPU, CPU->reg_b);
		cycles += 2;
		break;
	case 0x67:
		ea = ea_indexed(CPU, &cycles);
		r = inst_asr(CPU, read8(CPU, ea));
		write8(CPU, ea, (uint8_t)r);
		cycles += 6;
		break;
	case 0x77:
		ea = ea_extended(CPU);
		r = inst_asr(CPU, read8(CPU, ea));
		write8(CPU, ea, (uint8_t)r);
		cycles += 7;
		break;
		/* asl, asla, aslb */
	case 0x08:
		ea = ea_direct(CPU);
		r = inst_asl(CPU, read8(CPU, ea));
		write8(CPU, ea, (uint8_t)r);
		cycles += 6;
		break;
	case 0x48:
		CPU->reg_a = (uint8_t)inst_asl(CPU, CPU->reg_a);
		cycles += 2;
		break;
	case 0x58:
		CPU->reg_b = (uint8_t)inst_asl(CPU, CPU->reg_b);
		cycles += 2;
		break;
	case 0x68:
		ea = ea_indexed(CPU, &cycles);
		r = inst_asl(CPU, read8(CPU, ea));
		write8(CPU, ea, (uint8_t)r);
		cycles += 6;
		break;
	case 0x78:
		ea = ea_extended(CPU);
		r = inst_asl(CPU, read8(CPU, ea));
		write8(CPU, ea, (uint8_t)r);
		cycles += 7;
		break;
		/* rol, rola, rolb */
	case 0x09:
		ea = ea_direct(CPU);
		r = inst_rol(CPU, read8(CPU, ea));
		write8(CPU, ea, (uint8_t)r);
		cycles += 6;
		break;
	case 0x49:
		CPU->reg_a = (uint8_t)inst_rol(CPU, CPU->reg_a);
		cycles += 2;
		break;
	case 0x59:
		CPU->reg_b = (uint8_t)inst_rol(CPU, CPU->reg_b);
		cycles += 2;
		break;
	case 0x69:
		ea = ea_indexed(CPU, &cycles);
		r = inst_rol(CPU, read8(CPU, ea));
		write8(CPU, ea, (uint8_t)r);
		cycles += 6;
		break;
	case 0x79:
		ea = ea_extended(CPU);
		r = inst_rol(CPU, read8(CPU, ea));
		write8(CPU, ea, (uint8_t)r);
		cycles += 7;
		break;
		/* dec, deca, decb */
	case 0x0a:
		ea = ea_direct(CPU);
		r = inst_dec(CPU, read8(CPU, ea));
		write8(CPU, ea, (uint8_t)r);
		cycles += 6;
		break;
	case 0x4a:
		CPU->reg_a = (uint8_t)inst_dec(CPU, CPU->reg_a);
		cycles += 2;
		break;
	case 0x5a:
		CPU->reg_b = (uint8_t)inst_dec(CPU, CPU->reg_b);
		cycles += 2;
		break;
	case 0x6a:
		ea = ea_indexed(CPU, &cycles);
		r = inst_dec(CPU, read8(CPU, ea));
		write8(CPU, ea, (uint8_t)r);
		cycles += 6;
		break;
	case 0x7a:
		ea = ea_extended(CPU);
		r = inst_dec(CPU, read8(CPU, ea));
		write8(CPU, ea, (uint8_t)r);
		cycles += 7;
		break;
		/* inc, inca, incb */
	case 0x0c:
		ea = ea_direct(CPU);
		r = inst_inc(CPU, read8(CPU, ea));
		write8(CPU, ea, (uint8_t)r);
		cycles += 6;
		break;
	case 0x4c:
		CPU->reg_a = (uint8_t)inst_inc(CPU, CPU->reg_a);
		cycles += 2;
		break;
	case 0x5c:
		CPU->reg_b = (uint8_t)inst_inc(CPU, CPU->reg_b);
		cycles += 2;
		break;
	case 0x6c:
		ea = ea_indexed(CPU, &cycles);
		r = inst_inc(CPU, read8(CPU, ea));
		write8(CPU, ea, (uint8_t)r);
		cycles += 6;
		break;
	case 0x7c:
		ea = ea_extended(CPU);
		r = inst_inc(CPU, read8(CPU, ea));
		write8(CPU, ea, (uint8_t)r);
		cycles += 7;
		break;
		/* tst, tsta, tstb */
	case 0x0d:
		ea = ea_direct(CPU);
		inst_tst8(CPU, read8(CPU, ea));
		cycles += 6;
		break;
	case 0x4d:
		inst_tst8(CPU, CPU->reg_a);
		cycles += 2;
		break;
	case 0x5d:
		inst_tst8(CPU, CPU->reg_b);
		cycles += 2;
		break;
	case 0x6d:
		ea = ea_indexed(CPU, &cycles);
		inst_tst8(CPU, read8(CPU, ea));
		cycles += 6;
		break;
	case 0x7d:
		ea = ea_extended(CPU);
		inst_tst8(CPU, read8(CPU, ea));
		cycles += 7;
		break;
		/* jmp */
	case 0x0e:
		CPU->reg_pc = ea_direct(CPU);
		cycles += 3;
		break;
	case 0x6e:
		CPU->reg_pc = ea_indexed(CPU, &cycles);
		cycles += 3;
		break;
	case 0x7e:
		CPU->reg_pc = ea_extended(CPU);
		cycles += 4;
		break;
		/* clr */
	case 0x0f:
		ea = ea_direct(CPU);
		inst_clr(CPU);
		write8(CPU, ea, 0);
		cycles += 6;
		break;
	case 0x4f:
		inst_clr(CPU);
		CPU->reg_a = 0;
		cycles += 2;
		break;
	case 0x5f:
		inst_clr(CPU);
		CPU->reg_b = 0;
		cycles += 2;
		break;
	case 0x6f:
		ea = ea_indexed(CPU, &cycles);
		inst_clr(CPU);
		write8(CPU, ea, 0);
		cycles += 6;
		break;
	case 0x7f:
		ea = ea_extended(CPU);
		inst_clr(CPU);
		write8(CPU, ea, 0);
		cycles += 7;
		break;
		/* suba */
	case 0x80:
		CPU->reg_a = inst_sub8(CPU, CPU->reg_a, pc_read8(CPU));
		cycles += 2;
		break;
	case 0x90:
		ea = ea_direct(CPU);
		CPU->reg_a = inst_sub8(CPU, CPU->reg_a, read8(CPU, ea));
		cycles += 4;
		break;
	case 0xa0:
		ea = ea_indexed(CPU, &cycles);
		CPU->reg_a = inst_sub8(CPU, CPU->reg_a, read8(CPU, ea));
		cycles += 4;
		break;
	case 0xb0:
		ea = ea_extended(CPU);
		CPU->reg_a = inst_sub8(CPU, CPU->reg_a, read8(CPU, ea));
		cycles += 5;
		break;
		/* subb */
	case 0xc0:
		CPU->reg_b = inst_sub8(CPU, CPU->reg_b, pc_read8(CPU));
		cycles += 2;
		break;
	case 0xd0:
		ea = ea_direct(CPU);
		CPU->reg_b = inst_sub8(CPU, CPU->reg_b, read8(CPU, ea));
		cycles += 4;
		break;
	case 0xe0:
		ea = ea_indexed(CPU, &cycles);
		CPU->reg_b = inst_sub8(CPU, CPU->reg_b, read8(CPU, ea));
		cycles += 4;
		break;
	case 0xf0:
		ea = ea_extended(CPU);
		CPU->reg_b = inst_sub8(CPU, CPU->reg_b, read8(CPU, ea));
		cycles += 5;
		break;
		/* cmpa */
	case 0x81:
		inst_sub8(CPU, CPU->reg_a, pc_read8(CPU));
		cycles += 2;
		break;
	case 0x91:
		ea = ea_direct(CPU);
		inst_sub8(CPU, CPU->reg_a, read8(CPU, ea));
		cycles += 4;
		break;
	case 0xa1:
		ea = ea_indexed(CPU, &cycles);
		inst_sub8(CPU, CPU->reg_a, read8(CPU, ea));
		cycles += 4;
		break;
	case 0xb1:
		ea = ea_extended(CPU);
		inst_sub8(CPU, CPU->reg_a, read8(CPU, ea));
		cycles += 5;
		break;
		/* cmpb */
	case 0xc1:
		inst_sub8(CPU, CPU->reg_b, pc_read8(CPU));
		cycles += 2;
		break;
	case 0xd1:
		ea = ea_direct(CPU);
		inst_sub8(CPU, CPU->reg_b, read8(CPU, ea));
		cycles += 4;
		break;
	case 0xe1:
		ea = ea_indexed(CPU, &cycles);
		inst_sub8(CPU, CPU->reg_b, read8(CPU, ea));
		cycles += 4;
		break;
	case 0xf1:
		ea = ea_extended(CPU);
		inst_sub8(CPU, CPU->reg_b, read8(CPU, ea));
		cycles += 5;
		break;
		/* sbca */
	case 0x82:
		CPU->reg_a = inst_sbc(CPU, CPU->reg_a, pc_read8(CPU));
		cycles += 2;
		break;
	case 0x92:
		ea = ea_direct(CPU);
		CPU->reg_a = inst_sbc(CPU, CPU->reg_a, read8(CPU, ea));
		cycles += 4;
		break;
	case 0xa2:
		ea = ea_indexed(CPU, &cycles);
		CPU->reg_a = inst_sbc(CPU, CPU->reg_a, read8(CPU, ea));
		cycles += 4;
		break;
	case 0xb2:
		ea = ea_extended(CPU);
		CPU->reg_a = inst_sbc(CPU, CPU->reg_a, read8(CPU, ea));
		cycles += 5;
		break;
		/* sbcb */
	case 0xc2:
		CPU->reg_b = inst_sbc(CPU, CPU->reg_b, pc_read8(CPU));
		cycles += 2;
		break;
	case 0xd2:
		ea = ea_direct(CPU);
		CPU->reg_b = inst_sbc(CPU, CPU->reg_b, read8(CPU, ea));
		cycles += 4;
		break;
	case 0xe2:
		ea = ea_indexed(CPU, &cycles);
		CPU->reg_b = inst_sbc(CPU, CPU->reg_b, read8(CPU, ea));
		cycles += 4;
		break;
	case 0xf2:
		ea = ea_extended(CPU);
		CPU->reg_b = inst_sbc(CPU, CPU->reg_b, read8(CPU, ea));
		cycles += 5;
		break;
		/* anda */
	case 0x84:
		CPU->reg_a = inst_and(CPU, CPU->reg_a, pc_read8(CPU));
		cycles += 2;
		break;
	case 0x94:
		ea = ea_direct(CPU);
		CPU->reg_a = inst_and(CPU, CPU->reg_a, read8(CPU, ea));
		cycles += 4;
		break;
	case 0xa4:
		ea = ea_indexed(CPU, &cycles);
		CPU->reg_a = inst_and(CPU, CPU->reg_a, read8(CPU, ea));
		cycles += 4;
		break;
	case 0xb4:
		ea = ea_extended(CPU);
		CPU->reg_a = inst_and(CPU, CPU->reg_a, read8(CPU, ea));
		cycles += 5;
		break;
		/* andb */
	case 0xc4:
		CPU->reg_b = inst_and(CPU, CPU->reg_b, pc_read8(CPU));
		cycles += 2;
		break;
	case 0xd4:
		ea = ea_direct(CPU);
		CPU->reg_b = inst_and(CPU, CPU->reg_b, read8(CPU, ea));
		cycles += 4;
		break;
	case 0xe4:
		ea = ea_indexed(CPU, &cycles);
		CPU->reg_b = inst_and(CPU, CPU->reg_b, read8(CPU, ea));
		cycles += 4;
		break;
	case 0xf4:
		ea = ea_extended(CPU);
		CPU->reg_b = inst_and(CPU, CPU->reg_b, read8(CPU, ea));
		cycles += 5;
		break;
		/* bita */
	case 0x85:
		inst_and(CPU, CPU->reg_a, pc_read8(CPU));
		cycles += 2;
		break;
	case 0x95:
		ea = ea_direct(CPU);
		inst_and(CPU, CPU->reg_a, read8(CPU, ea));
		cycles += 4;
		break;
	case 0xa5:
		ea = ea_indexed(CPU, &cycles);
		inst_and(CPU, CPU->reg_a, read8(CPU, ea));
		cycles += 4;
		break;
	case 0xb5:
		ea = ea_extended(CPU);
		inst_and(CPU, CPU->reg_a, read8(CPU, ea));
		cycles += 5;
		break;
		/* bitb */
	case 0xc5:
		inst_and(CPU, CPU->reg_b, pc_read8(CPU));
		cycles += 2;
		break;
	case 0xd5:
		ea = ea_direct(CPU);
		inst_and(CPU, CPU->reg_b, read8(CPU, ea));
		cycles += 4;
		break;
	case 0xe5:
		ea = ea_indexed(CPU, &cycles);
		inst_and(CPU, CPU->reg_b, read8(CPU, ea));
		cycles += 4;
		break;
	case 0xf5:
		ea = ea_extended(CPU);
		inst_and(CPU, CPU->reg_b, read8(CPU, ea));
		cycles += 5;
		break;
		/* lda */
	case 0x86:
		CPU->reg_a = pc_read8(CPU);
		inst_tst8(CPU, CPU->reg_a);
		cycles += 2;
		break;
	case 0x96:
		ea = ea_direct(CPU);
		CPU->reg_a = read8(CPU, ea);
		inst_tst8(CPU, CPU->reg_a);
		cycles += 4;
		break;
	case 0xa6:
		ea = ea_indexed(CPU, &cycles);
		CPU->reg_a = read8(CPU, ea);
		inst_tst8(CPU, CPU->reg_a);
		cycles += 4;
		break;
	case 0xb6:
		ea = ea_extended(CPU);
		CPU->reg_a = read8(CPU, ea);
		inst_tst8(CPU, CPU->reg_a);
		cycles += 5;
		break;
		/* ldb */
	case 0xc6:
		CPU->reg_b = pc_read8(CPU);
		inst_tst8(CPU, CPU->reg_b);
		cycles += 2;
		break;
	case 0xd6:
		ea = ea_direct(CPU);
		CPU->reg_b = read8(CPU, ea);
		inst_tst8(CPU, CPU->reg_b);
		cycles += 4;
		break;
	case 0xe6:
		ea = ea_indexed(CPU, &cycles);
		CPU->reg_b = read8(CPU, ea);
		inst_tst8(CPU, CPU->reg_b);
		cycles += 4;
		break;
	case 0xf6:
		ea = ea_extended(CPU);
		CPU->reg_b = read8(CPU, ea);
		inst_tst8(CPU, CPU->reg_b);
		cycles += 5;
		break;
		/* sta */
	case 0x97:
		ea = ea_direct(CPU);
		write8(CPU, ea, CPU->reg_a);
		inst_tst8(CPU, CPU->reg_a);
		cycles += 4;
		break;
	case 0xa7:
		ea = ea_indexed(CPU, &cycles);
		write8(CPU, ea, CPU->reg_a);
		inst_tst8(CPU, CPU->reg_a);
		cycles += 4;
		break;
	case 0xb7:
		ea = ea_extended(CPU);
		write8(CPU, ea, CPU->reg_a);
		inst_tst8(CPU, CPU->reg_a);
		cycles += 5;
		break;
		/* stb */
	case 0xd7:
		ea = ea_direct(CPU);
		write8(CPU, ea, CPU->reg_b);
		inst_tst8(CPU, CPU->reg_b);
		cycles += 4;
		break;
	case 0xe7:
		ea = ea_indexed(CPU, &cycles);
		write8(CPU, ea, CPU->reg_b);
		inst_tst8(CPU, CPU->reg_b);
		cycles += 4;
		break;
	case 0xf7:
		ea = ea_extended(CPU);
		write8(CPU, ea, CPU->reg_b);
		inst_tst8(CPU, CPU->reg_b);
		cycles += 5;
		break;
		/* eora */
	case 0x88:
		CPU->reg_a = inst_eor(CPU, CPU->reg_a, pc_read8(CPU));
		cycles += 2;
		break;
	case 0x98:
		ea = ea_direct(CPU);
		CPU->reg_a = inst_eor(CPU, CPU->reg_a, read8(CPU, ea));
		cycles += 4;
		break;
	case 0xa8:
		ea = ea_indexed(CPU, &cycles);
		CPU->reg_a = inst_eor(CPU, CPU->reg_a, read8(CPU, ea));
		cycles += 4;
		break;
	case 0xb8:
		ea = ea_extended(CPU);
		CPU->reg_a = inst_eor(CPU, CPU->reg_a, read8(CPU, ea));
		cycles += 5;
		break;
		/* eorb */
	case 0xc8:
		CPU->reg_b = inst_eor(CPU, CPU->reg_b, pc_read8(CPU));
		cycles += 2;
		break;
	case 0xd8:
		ea = ea_direct(CPU);
		CPU->reg_b = inst_eor(CPU, CPU->reg_b, read8(CPU, ea));
		cycles += 4;
		break;
	case 0xe8:
		ea = ea_indexed(CPU, &cycles);
		CPU->reg_b = inst_eor(CPU, CPU->reg_b, read8(CPU, ea));
		cycles += 4;
		break;
	case 0xf8:
		ea = ea_extended(CPU);
		CPU->reg_b = inst_eor(CPU, CPU->reg_b, read8(CPU, ea));
		cycles += 5;
		break;
		/* adca */
	case 0x89:
		CPU->reg_a = inst_adc(CPU, CPU->reg_a, pc_read8(CPU));
		cycles += 2;
		break;
	case 0x99:
		ea = ea_direct(CPU);
		CPU->reg_a = inst_adc(CPU, CPU->reg_a, read8(CPU, ea));
		cycles += 4;
		break;
	case 0xa9:
		ea = ea_indexed(CPU, &cycles);
		CPU->reg_a = inst_adc(CPU, CPU->reg_a, read8(CPU, ea));
		cycles += 4;
		break;
	case 0xb9:
		ea = ea_extended(CPU);
		CPU->reg_a = inst_adc(CPU, CPU->reg_a, read8(CPU, ea));
		cycles += 5;
		break;
		/* adcb */
	case 0xc9:
		CPU->reg_b = inst_adc(CPU, CPU->reg_b, pc_read8(CPU));
		cycles += 2;
		break;
	case 0xd9:
		ea = ea_direct(CPU);
		CPU->reg_b = inst_adc(CPU, CPU->reg_b, read8(CPU, ea));
		cycles += 4;
		break;
	case 0xe9:
		ea = ea_indexed(CPU, &cycles);
		CPU->reg_b = inst_adc(CPU, CPU->reg_b, read8(CPU, ea));
		cycles += 4;
		break;
	case 0xf9:
		ea = ea_extended(CPU);
		CPU->reg_b = inst_adc(CPU, CPU->reg_b, read8(CPU, ea));
		cycles += 5;
		break;
		/* ora */
	case 0x8a:
		CPU->reg_a = inst_or(CPU, CPU->reg_a, pc_read8(CPU));
		cycles += 2;
		break;
	case 0x9a:
		ea = ea_direct(CPU);
		CPU->reg_a = inst_or(CPU, CPU->reg_a, read8(CPU, ea));
		cycles += 4;
		break;
	case 0xaa:
		ea = ea_indexed(CPU, &cycles);
		CPU->reg_a = inst_or(CPU, CPU->reg_a, read8(CPU, ea));
		cycles += 4;
		break;
	case 0xba:
		ea = ea_extended(CPU);
		CPU->reg_a = inst_or(CPU, CPU->reg_a, read8(CPU, ea));
		cycles += 5;
		break;
		/* orb */
	case 0xca:
		CPU->reg_b = inst_or(CPU, CPU->reg_b, pc_read8(CPU));
		cycles += 2;
		break;
	case 0xda:
		ea = ea_direct(CPU);
		CPU->reg_b = inst_or(CPU, CPU->reg_b, read8(CPU, ea));
		cycles += 4;
		break;
	case 0xea:
		ea = ea_indexed(CPU, &cycles);
		CPU->reg_b = inst_or(CPU, CPU->reg_b, read8(CPU, ea));
		cycles += 4;
		break;
	case 0xfa:
		ea = ea_extended(CPU);
		CPU->reg_b = inst_or(CPU, CPU->reg_b, read8(CPU, ea));
		cycles += 5;
		break;
		/* adda */
	case 0x8b:
		CPU->reg_a = inst_add8(CPU, CPU->reg_a, pc_read8(CPU));
		cycles += 2;
		break;
	case 0x9b:
		ea = ea_direct(CPU);
		CPU->reg_a = inst_add8(CPU, CPU->reg_a, read8(CPU, ea));
		cycles += 4;
		break;
	case 0xab:
		ea = ea_indexed(CPU, &cycles);
		CPU->reg_a = inst_add8(CPU, CPU->reg_a, read8(CPU, ea));
		cycles += 4;
		break;
	case 0xbb:
		ea = ea_extended(CPU);
		CPU->reg_a = inst_add8(CPU, CPU->reg_a, read8(CPU, ea));
		cycles += 5;
		break;
		/* addb */
	case 0xcb:
		CPU->reg_b = inst_add8(CPU, CPU->reg_b, pc_read8(CPU));
		cycles += 2;
		break;
	case 0xdb:
		ea = ea_direct(CPU);
		CPU->reg_b = inst_add8(CPU, CPU->reg_b, read8(CPU, ea));
		cycles += 4;
		break;
	case 0xeb:
		ea = ea_indexed(CPU, &cycles);
		CPU->reg_b = inst_add8(CPU, CPU->reg_b, read8(CPU, ea));
		cycles += 4;
		break;
	case 0xfb:
		ea = ea_extended(CPU);
		CPU->reg_b = inst_add8(CPU, CPU->reg_b, read8(CPU, ea));
		cycles += 5;
		break;
		/* subd */
	case 0x83:
		set_reg_d(CPU, inst_sub16(CPU, get_reg_d(CPU), pc_read16(CPU)));
		cycles += 4;
		break;
	case 0x93:
		ea = ea_direct(CPU);
		set_reg_d(CPU, inst_sub16(CPU, get_reg_d(CPU), read16(CPU, ea)));
		cycles += 6;
		break;
	case 0xa3:
		ea = ea_indexed(CPU, &cycles);
		set_reg_d(CPU, inst_sub16(CPU, get_reg_d(CPU), read16(CPU, ea)));
		cycles += 6;
		break;
	case 0xb3:
		ea = ea_extended(CPU);
		set_reg_d(CPU, inst_sub16(CPU, get_reg_d(CPU), read16(CPU, ea)));
		cycles += 7;
		break;
		/* cmpx */
	case 0x8c:
		inst_sub16(CPU, CPU->reg_x, pc_read16(CPU));
		cycles += 4;
		break;
	case 0x9c:
		ea = ea_direct(CPU);
		inst_sub16(CPU, CPU->reg_x, read16(CPU, ea));
		cycles += 6;
		break;
	case 0xac:
		ea = ea_indexed(CPU, &cycles);
		inst_sub16(CPU, CPU->reg_x, read16(CPU, ea));
		cycles += 6;
		break;
	case 0xbc:
		ea = ea_extended(CPU);
		inst_sub16(CPU, CPU->reg_x, read16(CPU, ea));
		cycles += 7;
		break;
		/* ldx */
	case 0x8e:
		CPU->reg_x = pc_read16(CPU);
		inst_tst16(CPU, CPU->reg_x);
		cycles += 3;
		break;
	case 0x9e:
		ea = ea_direct(CPU);
		CPU->reg_x = read16(CPU, ea);
		inst_tst16(CPU, CPU->reg_x);
		cycles += 5;
		break;
	case 0xae:
		ea = ea_indexed(CPU, &cycles);
		CPU->reg_x = read16(CPU, ea);
		inst_tst16(CPU, CPU->reg_x);
		cycles += 5;
		break;
	case 0xbe:
		ea = ea_extended(CPU);
		CPU->reg_x = read16(CPU, ea);
		inst_tst16(CPU, CPU->reg_x);
		cycles += 6;
		break;
		/* ldu */
	case 0xce:
		CPU->reg_u = pc_read16(CPU);
		inst_tst16(CPU, CPU->reg_u);
		cycles += 3;
		break;
	case 0xde:
		ea = ea_direct(CPU);
		CPU->reg_u = read16(CPU, ea);
		inst_tst16(CPU, CPU->reg_u);
		cycles += 5;
		break;
	case 0xee:
		ea = ea_indexed(CPU, &cycles);
		CPU->reg_u = read16(CPU, ea);
		inst_tst16(CPU, CPU->reg_u);
		cycles += 5;
		break;
	case 0xfe:
		ea = ea_extended(CPU);
		CPU->reg_u = read16(CPU, ea);
		inst_tst16(CPU, CPU->reg_u);
		cycles += 6;
		break;
		/* stx */
	case 0x9f:
		ea = ea_direct(CPU);
		write16(CPU, ea, CPU->reg_x);
		inst_tst16(CPU, CPU->reg_x);
		cycles += 5;
		break;
	case 0xaf:
		ea = ea_indexed(CPU, &cycles);
		write16(CPU, ea, CPU->reg_x);
		inst_tst16(CPU, CPU->reg_x);
		cycles += 5;
		break;
	case 0xbf:
		ea = ea_extended(CPU);
		write16(CPU, ea, CPU->reg_x);
		inst_tst16(CPU, CPU->reg_x);
		cycles += 6;
		break;
		/* stu */
	case 0xdf:
		ea = ea_direct(CPU);
		write16(CPU, ea, CPU->reg_u);
		inst_tst16(CPU, CPU->reg_u);
		cycles += 5;
		break;
	case 0xef:
		ea = ea_indexed(CPU, &cycles);
		write16(CPU, ea, CPU->reg_u);
		inst_tst16(CPU, CPU->reg_u);
		cycles += 5;
		break;
	case 0xff:
		ea = ea_extended(CPU);
		write16(CPU, ea, CPU->reg_u);
		inst_tst16(CPU, CPU->reg_u);
		cycles += 6;
		break;
		/* addd */
	case 0xc3:
		set_reg_d(CPU, inst_add16(CPU, get_reg_d(CPU), pc_read16(CPU)));
		cycles += 4;
		break;
	case 0xd3:
		ea = ea_direct(CPU);
		set_reg_d(CPU, inst_add16(CPU, get_reg_d(CPU), read16(CPU, ea)));
		cycles += 6;
		break;
	case 0xe3:
		ea = ea_indexed(CPU, &cycles);
		set_reg_d(CPU, inst_add16(CPU, get_reg_d(CPU), read16(CPU, ea)));
		cycles += 6;
		break;
	case 0xf3:
		ea = ea_extended(CPU);
		set_reg_d(CPU, inst_add16(CPU, get_reg_d(CPU), read16(CPU, ea)));
		cycles += 7;
		break;
		/* ldd */
	case 0xcc:
		set_reg_d(CPU, pc_read16(CPU));
		inst_tst16(CPU, get_reg_d(CPU));
		cycles += 3;
		break;
	case 0xdc:
		ea = ea_direct(CPU);
		set_reg_d(CPU, read16(CPU, ea));
		inst_tst16(CPU, get_reg_d(CPU));
		cycles += 5;
		break;
	case 0xec:
		ea = ea_indexed(CPU, &cycles);
		set_reg_d(CPU, read16(CPU, ea));
		inst_tst16(CPU, get_reg_d(CPU));
		cycles += 5;
		break;
	case 0xfc:
		ea = ea_extended(CPU);
		set_reg_d(CPU, read16(CPU, ea));
		inst_tst16(CPU, get_reg_d(CPU));
		cycles += 6;
		break;
		/* std */
	case 0xdd:
		ea = ea_direct(CPU);
		write16(CPU, ea, get_reg_d(CPU));
		inst_tst16(CPU, get_reg_d(CPU));
		cycles += 5;
		break;
	case 0xed:
		ea = ea_indexed(CPU, &cycles);
		write16(CPU, ea, get_reg_d(CPU));
		inst_tst16(CPU, get_reg_d(CPU));
		cycles += 5;
		break;
	case 0xfd:
		ea = ea_extended(CPU);
		write16(CPU, ea, get_reg_d(CPU));
		inst_tst16(CPU, get_reg_d(CPU));
		cycles += 6;
		break;
		/* nop */
	case 0x12:
		cycles += 2;
		break;
		/* mul */
	case 0x3d:
		r = (CPU->reg_a & 0xff) * (CPU->reg_b & 0xff);
		set_reg_d(CPU, r);

		set_cc(CPU, FLAG_Z, test_z16(r));
		set_cc(CPU, FLAG_C, (r >> 7) & 1);

		cycles += 11;
		break;
		/* bra */
	case 0x20:
		/* brn */
	case 0x21:
		inst_bra8(CPU, 0, op, &cycles);
		break;
		/* bhi */
	case 0x22:
		/* bls */
	case 0x23:
		inst_bra8(CPU, get_cc(CPU, FLAG_C) | get_cc(CPU, FLAG_Z), op, &cycles);
		break;
		/* bhs/bcc */
	case 0x24:
		/* blo/bcs */
	case 0x25:
		inst_bra8(CPU, get_cc(CPU, FLAG_C), op, &cycles);
		break;
		/* bne */
	case 0x26:
		/* beq */
	case 0x27:
		inst_bra8(CPU, get_cc(CPU, FLAG_Z), op, &cycles);
		break;
		/* bvc */
	case 0x28:
		/* bvs */
	case 0x29:
		inst_bra8(CPU, get_cc(CPU, FLAG_V), op, &cycles);
		break;
		/* bpl */
	case 0x2a:
		/* bmi */
	case 0x2b:
		inst_bra8(CPU, get_cc(CPU, FLAG_N), op, &cycles);
		break;
		/* bge */
	case 0x2c:
		/* blt */
	case 0x2d:
		inst_bra8(CPU, get_cc(CPU, FLAG_N) ^ get_cc(CPU, FLAG_V), op, &cycles);
		break;
		/* bgt */
	case 0x2e:
		/* ble */
	case 0x2f:
		inst_bra8(CPU, get_cc(CPU, FLAG_Z) |
			(get_cc(CPU, FLAG_N) ^ get_cc(CPU, FLAG_V)), op, &cycles);
		break;
		/* lbra */
	case 0x16:
		r = pc_read16(CPU);
		CPU->reg_pc += r;
		cycles += 5;
		break;
		/* lbsr */
	case 0x17:
		r = pc_read16(CPU);
		push16(CPU, &CPU->reg_s, CPU->reg_pc);
		CPU->reg_pc += r;
		cycles += 9;
		break;
		/* bsr */
	case 0x8d:
		r = pc_read8(CPU);
		push16(CPU, &CPU->reg_s, CPU->reg_pc);
		CPU->reg_pc += sign_extend(r);
		cycles += 7;
                if (CPU->reg_pc >= 0xD000) trace_bios(CPU);
		break;
		/* jsr */
	case 0x9d:
		ea = ea_direct(CPU);
		push16(CPU, &CPU->reg_s, CPU->reg_pc);
		CPU->reg_pc = ea;
		cycles += 7;
		break;
	case 0xad:
		ea = ea_indexed(CPU, &cycles);
		push16(CPU, &CPU->reg_s, CPU->reg_pc);
		CPU->reg_pc = ea;
		cycles += 7;
		break;
	case 0xbd:
		ea = ea_extended(CPU);
		push16(CPU, &CPU->reg_s, CPU->reg_pc);
		CPU->reg_pc = ea;
		cycles += 8;
		break;
		/* leax */
	case 0x30:
		CPU->reg_x = ea_indexed(CPU, &cycles);
		set_cc(CPU, FLAG_Z, test_z16(CPU->reg_x));
		cycles += 4;
		break;
		/* leay */
	case 0x31:
		CPU->reg_y = ea_indexed(CPU, &cycles);
		set_cc(CPU, FLAG_Z, test_z16(CPU->reg_y));
		cycles += 4;
		break;
		/* leas */
	case 0x32:
		CPU->reg_s = ea_indexed(CPU, &cycles);
		cycles += 4;
		break;
		/* leau */
	case 0x33:
		CPU->reg_u = ea_indexed(CPU, &cycles);
		cycles += 4;
		break;
		/* pshs */
	case 0x34:
		inst_psh(CPU, pc_read8(CPU), &CPU->reg_s, CPU->reg_u, &cycles);
		cycles += 5;
		break;
		/* puls */
	case 0x35:
		inst_pul(CPU, pc_read8(CPU), &CPU->reg_s, &CPU->reg_u, &cycles);
		cycles += 5;
		break;
		/* pshu */
	case 0x36:
		inst_psh(CPU, pc_read8(CPU), &CPU->reg_u, CPU->reg_s, &cycles);
		cycles += 5;
		break;
		/* pulu */
	case 0x37:
		inst_pul(CPU, pc_read8(CPU), &CPU->reg_u, &CPU->reg_s, &cycles);
		cycles += 5;
		break;
		/* rts */
	case 0x39:
		CPU->reg_pc = pull16(CPU, &CPU->reg_s);
		cycles += 5;
		break;
		/* abx */
	case 0x3a:
		CPU->reg_x += CPU->reg_b & 0xff;
		cycles += 3;
		break;
		/* orcc */
	case 0x1a:
		CPU->reg_cc |= pc_read8(CPU);
		cycles += 3;
		break;
		/* andcc */
	case 0x1c:
		CPU->reg_cc &= pc_read8(CPU);
		cycles += 3;
		break;
		/* sex */
	case 0x1d:
		set_reg_d(CPU, sign_extend(CPU->reg_b));
		set_cc(CPU, FLAG_N, test_n(CPU->reg_a));
		set_cc(CPU, FLAG_Z, test_z16(get_reg_d(CPU)));
		cycles += 2;
		break;
		/* exg */
	case 0x1e:
		inst_exg(CPU);
		cycles += 8;
		break;
		/* tfr */
	case 0x1f:
		inst_tfr(CPU);
		cycles += 6;
		break;
		/* rti */
	case 0x3b:
		if (get_cc(CPU, FLAG_E))
		{
			inst_pul(CPU, 0xff, &CPU->reg_s, &CPU->reg_u, &cycles);
		}
		else
		{
			inst_pul(CPU, 0x81, &CPU->reg_s, &CPU->reg_u, &cycles);
		}

		cycles += 3;
		break;
		/* swi */
	case 0x3f:
		set_cc(CPU, FLAG_E, 1);
		inst_psh(CPU, 0xff, &CPU->reg_s, CPU->reg_u, &cycles);
		set_cc(CPU, FLAG_I, 1);
		set_cc(CPU, FLAG_F, 1);
		CPU->reg_pc = read16(CPU, 0xfffa);
		cycles += 7;
		break;
		/* sync */
	case 0x13:
		CPU->irq_status = IRQ_SYNC;
		cycles += 2;
		break;
		/* daa */
	case 0x19:
		i0 = CPU->reg_a;
		i1 = 0;

		if ((CPU->reg_a & 0x0f) > 0x09 || get_cc(CPU, FLAG_H) == 1)
		{
			i1 |= 0x06;
		}

		if ((CPU->reg_a & 0xf0) > 0x80 && (CPU->reg_a & 0x0f) > 0x09)
		{
			i1 |= 0x60;
		}

		if ((CPU->reg_a & 0xf0) > 0x90 || get_cc(CPU, FLAG_C) == 1)
		{
			i1 |= 0x60;
		}

		CPU->reg_a = i0 + i1;

		set_cc(CPU, FLAG_N, test_n(CPU->reg_a));
		set_cc(CPU, FLAG_Z, test_z8(CPU->reg_a));
		set_cc(CPU, FLAG_V, 0);
		set_cc(CPU, FLAG_C, test_c(i0, i1, CPU->reg_a, 0));
		cycles += 2;
		break;
		/* cwai */
	case 0x3c:
		CPU->reg_cc &= pc_read8(CPU);
		set_cc(CPU, FLAG_E, 1);
		inst_psh(CPU, 0xff, &CPU->reg_s, CPU->reg_u, &cycles);
		CPU->irq_status = IRQ_CWAI;
		cycles += 4;
		break;

		/* page 1 instructions */

	case 0x10:
		op = pc_read8(CPU);

		switch (op)
		{
			/* lbra */
		case 0x20:
			/* lbrn */
		case 0x21:
			inst_bra16(CPU, 0, op, &cycles);
			break;
			/* lbhi */
		case 0x22:
			/* lbls */
		case 0x23:
			inst_bra16(CPU, get_cc(CPU, FLAG_C) | get_cc(CPU, FLAG_Z), op, &cycles);
			break;
			/* lbhs/lbcc */
		case 0x24:
			/* lblo/lbcs */
		case 0x25:
			inst_bra16(CPU, get_cc(CPU, FLAG_C), op, &cycles);
			break;
			/* lbne */
		case 0x26:
			/* lbeq */
		case 0x27:
			inst_bra16(CPU, get_cc(CPU, FLAG_Z), op, &cycles);
			break;
			/* lbvc */
		case 0x28:
			/* lbvs */
		case 0x29:
			inst_bra16(CPU, get_cc(CPU, FLAG_V), op, &cycles);
			break;
			/* lbpl */
		case 0x2a:
			/* lbmi */
		case 0x2b:
			inst_bra16(CPU, get_cc(CPU, FLAG_N), op, &cycles);
			break;
			/* lbge */
		case 0x2c:
			/* lblt */
		case 0x2d:
			inst_bra16(CPU, get_cc(CPU, FLAG_N) ^ get_cc(CPU, FLAG_V), op, &cycles);
			break;
			/* lbgt */
		case 0x2e:
			/* lble */
		case 0x2f:
			inst_bra16(CPU, get_cc(CPU, FLAG_Z) |
				(get_cc(CPU, FLAG_N) ^ get_cc(CPU, FLAG_V)), op, &cycles);
			break;
			/* cmpd */
		case 0x83:
			inst_sub16(CPU, get_reg_d(CPU), pc_read16(CPU));
			cycles += 5;
			break;
		case 0x93:
			ea = ea_direct(CPU);
			inst_sub16(CPU, get_reg_d(CPU), read16(CPU, ea));
			cycles += 7;
			break;
		case 0xa3:
			ea = ea_indexed(CPU, &cycles);
			inst_sub16(CPU, get_reg_d(CPU), read16(CPU, ea));
			cycles += 7;
			break;
		case 0xb3:
			ea = ea_extended(CPU);
			inst_sub16(CPU, get_reg_d(CPU), read16(CPU, ea));
			cycles += 8;
			break;
			/* cmpy */
		case 0x8c:
			inst_sub16(CPU, CPU->reg_y, pc_read16(CPU));
			cycles += 5;
			break;
		case 0x9c:
			ea = ea_direct(CPU);
			inst_sub16(CPU, CPU->reg_y, read16(CPU, ea));
			cycles += 7;
			break;
		case 0xac:
			ea = ea_indexed(CPU, &cycles);
			inst_sub16(CPU, CPU->reg_y, read16(CPU, ea));
			cycles += 7;
			break;
		case 0xbc:
			ea = ea_extended(CPU);
			inst_sub16(CPU, CPU->reg_y, read16(CPU, ea));
			cycles += 8;
			break;
			/* ldy */
		case 0x8e:
			CPU->reg_y = pc_read16(CPU);
			inst_tst16(CPU, CPU->reg_y);
			cycles += 4;
			break;
		case 0x9e:
			ea = ea_direct(CPU);
			CPU->reg_y = read16(CPU, ea);
			inst_tst16(CPU, CPU->reg_y);
			cycles += 6;
			break;
		case 0xae:
			ea = ea_indexed(CPU, &cycles);
			CPU->reg_y = read16(CPU, ea);
			inst_tst16(CPU, CPU->reg_y);
			cycles += 6;
			break;
		case 0xbe:
			ea = ea_extended(CPU);
			CPU->reg_y = read16(CPU, ea);
			inst_tst16(CPU, CPU->reg_y);
			cycles += 7;
			break;
			/* sty */
		case 0x9f:
			ea = ea_direct(CPU);
			write16(CPU, ea, CPU->reg_y);
			inst_tst16(CPU, CPU->reg_y);
			cycles += 6;
			break;
		case 0xaf:
			ea = ea_indexed(CPU, &cycles);
			write16(CPU, ea, CPU->reg_y);
			inst_tst16(CPU, CPU->reg_y);
			cycles += 6;
			break;
		case 0xbf:
			ea = ea_extended(CPU);
			write16(CPU, ea, CPU->reg_y);
			inst_tst16(CPU, CPU->reg_y);
			cycles += 7;
			break;
			/* lds */
		case 0xce:
			CPU->reg_s = pc_read16(CPU);
			inst_tst16(CPU, CPU->reg_s);
			cycles += 4;
			break;
		case 0xde:
			ea = ea_direct(CPU);
			CPU->reg_s = read16(CPU, ea);
			inst_tst16(CPU, CPU->reg_s);
			cycles += 6;
			break;
		case 0xee:
			ea = ea_indexed(CPU, &cycles);
			CPU->reg_s = read16(CPU, ea);
			inst_tst16(CPU, CPU->reg_s);
			cycles += 6;
			break;
		case 0xfe:
			ea = ea_extended(CPU);
			CPU->reg_s = read16(CPU, ea);
			inst_tst16(CPU, CPU->reg_s);
			cycles += 7;
			break;
			/* sts */
		case 0xdf:
			ea = ea_direct(CPU);
			write16(CPU, ea, CPU->reg_s);
			inst_tst16(CPU, CPU->reg_s);
			cycles += 6;
			break;
		case 0xef:
			ea = ea_indexed(CPU, &cycles);
			write16(CPU, ea, CPU->reg_s);
			inst_tst16(CPU, CPU->reg_s);
			cycles += 6;
			break;
		case 0xff:
			ea = ea_extended(CPU);
			write16(CPU, ea, CPU->reg_s);
			inst_tst16(CPU, CPU->reg_s);
			cycles += 7;
			break;
			/* swi2 */
		case 0x3f:
			set_cc(CPU, FLAG_E, 1);
			inst_psh(CPU, 0xff, &CPU->reg_s, CPU->reg_u, &cycles);
			CPU->reg_pc = read16(CPU, 0xfff4);
			cycles += 8;
			break;
		default:
			printf("unknown page-1 op code: %.2x\n", op);
			break;
		}

		break;

		/* page 2 instructions */

	case 0x11:
		op = pc_read8(CPU);

		switch (op)
		{
			/* cmpu */
		case 0x83:
			inst_sub16(CPU, CPU->reg_u, pc_read16(CPU));
			cycles += 5;
			break;
		case 0x93:
			ea = ea_direct(CPU);
			inst_sub16(CPU, CPU->reg_u, read16(CPU, ea));
			cycles += 7;
			break;
		case 0xa3:
			ea = ea_indexed(CPU, &cycles);
			inst_sub16(CPU, CPU->reg_u, read16(CPU, ea));
			cycles += 7;
			break;
		case 0xb3:
			ea = ea_extended(CPU);
			inst_sub16(CPU, CPU->reg_u, read16(CPU, ea));
			cycles += 8;
			break;
			/* cmps */
		case 0x8c:
			inst_sub16(CPU, CPU->reg_s, pc_read16(CPU));
			cycles += 5;
			break;
		case 0x9c:
			ea = ea_direct(CPU);
			inst_sub16(CPU, CPU->reg_s, read16(CPU, ea));
			cycles += 7;
			break;
		case 0xac:
			ea = ea_indexed(CPU, &cycles);
			inst_sub16(CPU, CPU->reg_s, read16(CPU, ea));
			cycles += 7;
			break;
		case 0xbc:
			ea = ea_extended(CPU);
			inst_sub16(CPU, CPU->reg_s, read16(CPU, ea));
			cycles += 8;
			break;
			/* swi3 */
		case 0x3f:
			set_cc(CPU, FLAG_E, 1);
			inst_psh(CPU, 0xff, &CPU->reg_s, CPU->reg_u, &cycles);
			CPU->reg_pc = read16(CPU, 0xfff2);
			cycles += 8;
			break;
		default:
			printf("unknown page-2 op code: %.2x\n", op);
			break;
		}

		break;

	default:
		printf("unknown page-0 op code: %.2x\n", op);
		break;
	}

	return cycles;
}
