#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#ifdef Vectrex_SDL
#include <SDL.h>
#include <SDL_image.h>
#endif
#include "e6809.h"
#include "e8910.h"
#include "e6522.h"
#include "edac.h"
#include "vecx.h"
#include "ser.h"

enum
{
	EMU_TIMER = 20, /* the emulators heart beats at 20 milliseconds */

	DEFAULT_WIDTH = 495,
	DEFAULT_HEIGHT = 615
};

vecx vectrex;
#ifdef Vectrex_SDL
static SDL_Window *window = NULL;
static SDL_Renderer *renderer = NULL;
static SDL_Texture *overlay = NULL;
static SDL_Texture *buffer = NULL;
static SDL_Texture *buffer2 = NULL;

static int32_t scl_factor;
#endif

static void quit(void);

/* command line arguments */
static char *bios_filename = "bios.bin";
static char *cart_filename = NULL;
static char *overlay_filename = NULL;
static char fullscreen = 0;

static void render(void)
{
#ifdef Vectrex_SDL
	SDL_SetRenderTarget(renderer, buffer);
	{
		SDL_SetRenderDrawColor(renderer, 0, 0, 0, 128);
		SDL_RenderFillRect(renderer, NULL);

		for (size_t v = 0; v < vectrex.vector_draw_cnt; v++)
		{
			Uint8 c = vectrex.vectors[v].color * 256 / VECTREX_COLORS;
			int x0 = vectrex.vectors[v].x0 / scl_factor;
			int y0 = vectrex.vectors[v].y0 / scl_factor;
			int x1 = vectrex.vectors[v].x1 / scl_factor;
			int y1 = vectrex.vectors[v].y1 / scl_factor;

			SDL_SetRenderDrawColor(renderer, 255, 255, 255, c);
			if (x0 == x1 && y0 == y1)
			{
				/* point */
				SDL_RenderDrawPoint(renderer, x0, y0);
				SDL_RenderDrawPoint(renderer, x0 + 1, y0);
				SDL_RenderDrawPoint(renderer, x0, y0 + 1);
				SDL_RenderDrawPoint(renderer, x0 + 1, y0 + 1);
			}
			else
			{
				SDL_RenderDrawLine(renderer, x0, y0, x1, y1);
				SDL_RenderDrawLine(renderer, x0 + 1, y0 + 1, x1 + 1, y1 + 1);
			}
		}
	}

	SDL_SetRenderTarget(renderer, buffer2);
	{
		SDL_RenderCopy(renderer, buffer, NULL, NULL);
	}

	SDL_SetRenderTarget(renderer, NULL);
	{
		SDL_SetRenderDrawColor(renderer, 0, 0, 0, 255);
		SDL_RenderClear(renderer);

		SDL_RenderCopy(renderer, buffer, NULL, NULL);
		SDL_RenderCopy(renderer, buffer2, NULL, NULL);

		if (overlay)
		{
			SDL_RenderCopy(renderer, overlay, NULL, NULL);
		}
	}
	SDL_RenderPresent(renderer);
#endif
}

static void load_bios(void)
{
	FILE *f;
	if (!(f = fopen(bios_filename, "rb")))
	{
		perror(bios_filename);
		quit();
	}
	if (fread(vectrex.rom, 1, 8192/*sizeof(vectrex.rom)*/, f) != 8192 /*sizeof(vectrex.rom)*/)
	{
		fprintf(stderr, "Invalid bios length\n");
		quit();
	}
	fclose(f);
}

static void load_cart(void)
{
        // memset(vectrex.cart, 0, sizeof(vectrex.cart));
	if (cart_filename)
	{
		FILE *f;
		if (!(f = fopen(cart_filename, "rb")))
		{
			perror(cart_filename);
			return;
		}
		(void)fread(vectrex.cart, 1, 32768 /*sizeof(vectrex.cart)*/, f);
		fclose(f);
	}
}

static void resize(void)
{
#ifdef Vectrex_SDL
	int sclx, scly;
	int screenx, screeny;
	int width, height;

	SDL_GetWindowSize(window, &screenx, &screeny);

	sclx = DAC_MAX_X / screenx;
	scly = DAC_MAX_Y / screeny;

	scl_factor = sclx > scly ? sclx : scly;
	width = DAC_MAX_X / scl_factor;
	height = DAC_MAX_Y / scl_factor;

	SDL_RenderSetLogicalSize(renderer, width, height);

	SDL_DestroyTexture(buffer);
	SDL_DestroyTexture(buffer2);

	buffer = SDL_CreateTexture(renderer, SDL_PIXELFORMAT_RGBA8888, SDL_TEXTUREACCESS_TARGET, width, height);
	buffer2 = SDL_CreateTexture(renderer, SDL_PIXELFORMAT_RGBA8888, SDL_TEXTUREACCESS_TARGET, width / 2, height / 2);

	SDL_SetTextureBlendMode(buffer2, SDL_BLENDMODE_BLEND);
	SDL_SetTextureAlphaMod(buffer2, 128);
#endif
}

static int readevents(void)
{
#ifdef Vectrex_SDL
	SDL_Event e;
	while (SDL_PollEvent(&e))
	{
		switch (e.type)
		{
		case SDL_QUIT:
			return 1;
			break;
		case SDL_WINDOWEVENT:
			if (e.window.event == SDL_WINDOWEVENT_RESIZED)
				resize();
			break;
		case SDL_DROPFILE:
			cart_filename = e.drop.file;
			load_cart();
			vecx_reset(&vectrex);
			break;
		case SDL_KEYDOWN:
			switch (e.key.keysym.sym)
			{
			case SDLK_ESCAPE: return 1;
			case SDLK_a: vecx_input(&vectrex, VECTREX_PAD1_BUTTON1, 1); break;
			case SDLK_s: vecx_input(&vectrex, VECTREX_PAD1_BUTTON2, 1); break;
			case SDLK_d: vecx_input(&vectrex, VECTREX_PAD1_BUTTON3, 1); break;
			case SDLK_f: vecx_input(&vectrex, VECTREX_PAD1_BUTTON4, 1); break;
			case SDLK_LEFT: vecx_input(&vectrex, VECTREX_PAD1_X, 0x00); break;
			case SDLK_RIGHT: vecx_input(&vectrex, VECTREX_PAD1_X, 0xff); break;
			case SDLK_UP: vecx_input(&vectrex, VECTREX_PAD1_Y, 0xff); break;
			case SDLK_DOWN: vecx_input(&vectrex, VECTREX_PAD1_Y, 0x00); break;
			}
			break;
		case SDL_KEYUP:
			switch (e.key.keysym.sym)
			{
			case SDLK_F1: vecx_load(&vectrex, "q.save"); break;
			case SDLK_F2: vecx_save(&vectrex, "q.save"); break;
			case SDLK_r: load_cart(); vecx_reset(&vectrex); break;

			case SDLK_a: vecx_input(&vectrex, VECTREX_PAD1_BUTTON1, 0); break;
			case SDLK_s: vecx_input(&vectrex, VECTREX_PAD1_BUTTON2, 0); break;
			case SDLK_d: vecx_input(&vectrex, VECTREX_PAD1_BUTTON3, 0); break;
			case SDLK_f: vecx_input(&vectrex, VECTREX_PAD1_BUTTON4, 0); break;
			case SDLK_LEFT: vecx_input(&vectrex, VECTREX_PAD1_X, 0x80); break;
			case SDLK_RIGHT: vecx_input(&vectrex, VECTREX_PAD1_X, 0x80); break;
			case SDLK_UP: vecx_input(&vectrex, VECTREX_PAD1_Y, 0x80); break;
			case SDLK_DOWN: vecx_input(&vectrex, VECTREX_PAD1_Y, 0x80); break;
			}
			break;
		}
	}
#endif
	return 0;
}

static void emuloop(void)
{
#ifdef Vectrex_SDL
	Uint32 next_time = SDL_GetTicks() + EMU_TIMER;
#endif
	vecx_reset(&vectrex);
	for (;;)
	{
		vecx_emu(&vectrex, (VECTREX_MHZ / 1000) * EMU_TIMER);
		if (readevents()) break;

#ifdef Vectrex_SDL
		{
			Uint32 now = SDL_GetTicks();
			if (now < next_time) {
                          SDL_Delay(next_time - now);
			} else
				next_time = now;
			next_time += EMU_TIMER;
		}
#endif
	}
}

static void load_overlay()
{
#ifdef Vectrex_SDL
	if (overlay_filename)
	{
		SDL_Texture *image = IMG_LoadTexture(renderer, overlay_filename);
		if (image)
		{
			overlay = image;
			SDL_SetTextureBlendMode(image, SDL_BLENDMODE_BLEND);
			SDL_SetTextureAlphaMod(image, 128);
		}
		else
		{
			fprintf(stderr, "IMG_Load: %s\n", IMG_GetError());
		}
	}
#endif
}

static int init(void)
{
#ifdef Vectrex_SDL
        if (SDL_Init(SDL_INIT_VIDEO /* | SDL_INIT_AUDIO */) < 0) // turn off sound while developing
	{
		fprintf(stderr, "Failed to initialize SDL: %s\n", SDL_GetError());
		return 0;
	}

	window = SDL_CreateWindow("Vecx", SDL_WINDOWPOS_CENTERED, SDL_WINDOWPOS_CENTERED, DEFAULT_WIDTH, DEFAULT_HEIGHT, SDL_WINDOW_RESIZABLE);
	if (!window)
	{
		fprintf(stderr, "Failed to create window: %s\n", SDL_GetError());
		return 0;
	}

	renderer = SDL_CreateRenderer(window, -1, SDL_RENDERER_ACCELERATED | SDL_RENDERER_TARGETTEXTURE);
	if (!renderer)
	{
		fprintf(stderr, "Failed to create renderer: %s\n", SDL_GetError());
		return 0;
	}

	SDL_SetRenderDrawBlendMode(renderer, SDL_BLENDMODE_BLEND);
	SDL_SetHint(SDL_HINT_RENDER_SCALE_QUALITY, "1");

	if (fullscreen)
		SDL_SetWindowFullscreen(window, SDL_WINDOW_FULLSCREEN_DESKTOP);
#endif
	return 1;
}

static void quit(void)
{
#ifdef Vectrex_SDL
  // Sometimes the program appears to hang when you press ESCape ... still to be debugged.
  //  fprintf(stderr, "Quit #1\n");
        if (renderer)
		SDL_DestroyRenderer(renderer);
  //  fprintf(stderr, "Quit #2\n");
	if (window)
		SDL_DestroyWindow(window);
  //  fprintf(stderr, "Quit #3\n");
	SDL_Quit();
  //  fprintf(stderr, "Quit #4\n");
#endif
        // Write the disassembly file
        // Still TO DO: Initialise these arrays by loading the same file at startup,
        // so that the file accumulates more data over time, rather than just reflecting
        // the most recent run.
        {
          int i, dup;
          FILE *dat;
          char *datfile = malloc(strlen(cart_filename)+16);
          sprintf(datfile, "%s.inf", cart_filename);
          dat = fopen(datfile, "w");
  //  fprintf(stderr, "Quit #5\n");
          if (dat == NULL) {
            int fd = mkstemp("/tmp/tmpXXXXXX");
            //sprintf(datfile, "%s", tmpnam(NULL));
            dat = fdopen(fd, "w");
            // if (fd >= 0) close(fd);
            if (dat == NULL) {
              fprintf(stderr, "ERROR: vecx cannot open %s.inf OR %s to save dat information\n", cart_filename, datfile);
              exit(1);
            }
            fprintf(stderr, "WARNING: vecx cannot open %s.inf to save dat information - saving to %s instead\n", cart_filename, datfile);
          }
  //  fprintf(stderr, "Quit #6\n");
          fprintf(dat, "V %X C %s\n", DatVersion, cart_filename);
  //  fprintf(stderr, "Quit #7\n");
          for (i = 0; i <= 0xFFFF; i++) {
            fprintf(dat, "%04X F %X ", i, GLOBAL_vecx->CPU.F[i]); // mark the ways in which ram/rom addresses are used.
            fprintf(dat, "D %X ", GLOBAL_vecx->CPU.DP[i]); // Note the value of DP when executing code at this address. 100 is unassigned 200 is more than one value seen.
            fprintf(dat, "B %X ", GLOBAL_vecx->CPU.IndexBase[i]); // When loading table,x for any address, save the value of 'table' for this address.
            fprintf(dat, "I %X", GLOBAL_vecx->CPU.Index[i]);     // When loading table,x for any address, save the value of 'x' for this address.
            dup = 0;
            while ((i < 0xFFFF)
                   && (GLOBAL_vecx->CPU.F[i] == GLOBAL_vecx->CPU.F[i+1])
                   && (GLOBAL_vecx->CPU.DP[i] == GLOBAL_vecx->CPU.DP[i+1])
                   && (GLOBAL_vecx->CPU.IndexBase[i] == GLOBAL_vecx->CPU.IndexBase[i+1])
                   && (GLOBAL_vecx->CPU.Index[i] == GLOBAL_vecx->CPU.Index[i+1])) {
              dup += 1; i += 1;
            }
            fprintf(dat, " %d\n", dup);
          }
  //  fprintf(stderr, "Quit #8\n");
        }
        
  //  fprintf(stderr, "Quit #9\n");
	exit(0);
}

void parse_args(int argc, char* argv[])
{
	for (int i = 1; i < argc; i++)
	{
		if (strcmp(argv[i], "--help") == 0 || strcmp(argv[i], "-h") == 0)
		{
			puts("Usage: vecx [options] [cart_file]");
			puts("Options:");
			puts("  --help            Display this help message");
			puts("  --bios <file>     Load bios file");
#ifdef Vectrex_SDL
			puts("  --overlay <file>  Load overlay file");
			puts("  --fullscreen      Launch in fullscreen mode");
#endif
			exit(0);
		}
		else if (strcmp(argv[i], "--bios") == 0 || strcmp(argv[i], "-b") == 0)
		{
			bios_filename = argv[++i];
		}
#ifdef Vectrex_SDL
		else if (strcmp(argv[i], "--overlay") == 0 || strcmp(argv[i], "-o") == 0)
		{
			overlay_filename = argv[++i];
		}
		else if (strcmp(argv[i], "--fullscreen") == 0 || strcmp(argv[i], "-f") == 0)
		{
			fullscreen = 1;
		}
#endif
		else if (i == argc - 1)
		{
			cart_filename = argv[i];
		}
		else
		{
			printf("Unkown flag: %s\n", argv[i]);
			exit(0);
		}
	}
}

int main(int argc, char *argv[])
{
  int i;

  for (i = 0; i <= 0xFFFF; i++) {
    vectrex.memory[i] = 0;
    vectrex.disasm[i] = NULL;
  }
  vectrex.rom  = &vectrex.memory[0xE000]; //[8192];
  vectrex.cart = &vectrex.memory[0x0000]; //[32768];
  vectrex.ram  = &vectrex.memory[0xC800]; //[1024];

  vectrex.CPU.memory = vectrex.memory;
  vectrex.CPU.disasm = vectrex.disasm;
  
  parse_args(argc, argv);

  if (!init())
    quit();

  resize();
  load_bios();
  load_cart();
  load_overlay();
  e8910_init(&vectrex.PSG);
  vectrex.render = render;

  // if resuming last play, 'vecx_load(&vectrex, "q.save");' but with cartridge name

  emuloop();

  // to save state for resuming, 'vecx_save(&vectrex, "q.save");' but with cartridge name

  e8910_done(&vectrex.PSG);

  quit();

  return 0;
}
