#include <errno.h>
#include <error.h>
#include <fcntl.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>

#include "vecx.h"
extern int Data_From_USB_Fd, Data_To_USB_Fd;
extern int serial_port;

#include "e6522.h"
#include "e6809.h"
#include "e8910.h"
#include "edac.h"

char *q = NULL;
int Q_SIZE = 4096;
int q_get = 0;
int q_put = 0;

static int orig_flags;
static struct termios orig_termios;

void restore_console(void) {
  // tcsetattr(0, TCSANOW, &orig_termios);
  if (tcsetattr(Data_From_USB_Fd, TCSAFLUSH, &orig_termios) < 0) {
    fprintf(stderr, "can't restore terminal mode\n");
    exit(4);
  }
  fcntl(Data_From_USB_Fd, F_SETFL, orig_flags);
  if (serial_port != 4) {
    fprintf(stderr, "\rRestoring terminal settings\r\n\n");
    fflush(stderr);
  } else {
    fprintf(stderr, "\r\n");
    fflush(stderr);
    system("stty sane");
  }
}

void init_console(void) {
  struct termios t;
  int flags = fcntl(Data_From_USB_Fd, F_GETFL, 0);
  orig_flags = flags;
  fcntl(Data_From_USB_Fd, F_SETFL, flags | O_NONBLOCK); // put USB into non-blocking mode
  tcgetattr(Data_From_USB_Fd, &t);
  tcgetattr(Data_From_USB_Fd, &orig_termios);
  // t.c_lflag &= ~(ICANON|ECHO);

  /* input modes - clear indicated ones giving: no break, no CR to NL,
     no parity check, no strip char, no start/stop output (sic) control */
  t.c_iflag &= ~(BRKINT | ICRNL | INPCK | ISTRIP | IXON);

  /* output modes - clear giving: no post processing such as NL to CR+NL */
  t.c_oflag &= ~(OPOST);

  /* control modes - set 8 bit chars */
  t.c_cflag |= (CS8);

  /* local modes - clear giving: echoing off, canonical off (no erase with
     backspace, ^U,...),  no extended functions, no signal chars (^Z,^C) */
  t.c_lflag &= ~(ECHO | ICANON | IEXTEN | ISIG);

#ifdef NEVER
  /* control chars - set return condition: min number of bytes and timer */
  t.c_cc[VMIN] = 5;
  t.c_cc[VTIME] = 8; /* after 5 bytes or .8 seconds
                            after first byte seen      */
  t.c_cc[VMIN] = 0;
  t.c_cc[VTIME] = 0; /* immediate - anything       */
  t.c_cc[VMIN] = 2;
  t.c_cc[VTIME] = 0; /* after two bytes, no timer  */
  t.c_cc[VMIN] = 0;
  t.c_cc[VTIME] = 8; /* after a byte or .8 seconds */
#endif
  // tcsetattr(0, TCSANOW, &t);
  if (tcsetattr(Data_From_USB_Fd, TCSAFLUSH, &t) < 0) {
    fprintf(stderr, "can't set raw mode\n");
    exit(3);
  }
  /* put terminal in raw mode after flushing */
  atexit(restore_console);
  /*
       Functions registered using atexit() (and on_exit(3)) are not
       called if a process terminates abnormally because of the delivery
       of a signal.
   */
}

void q_push(char c) {
  if (q == NULL) {
    q = malloc(sizeof(char) * Q_SIZE);
    q_get = q_put = 0;
  }
  q[q_put] = c;
  q_put = (q_put + 1) % Q_SIZE;
  if (q_get == q_put) {
    // FULL! Expand it.
    fprintf(stderr, "Input ACIA queue is full!\n");
    exit(2);
  }
}

char q_pop(void) {
  int c;
  if (q_put != q_get) {
    c = q[q_get];
    q_get = (q_get + 1) % Q_SIZE;
    if (c == '\n') c = '\r';
    return c;
  } else {
    // probably a programmer error.
    // Looks like this came from a TST ,X instruction?
    // fprintf(stderr, "q_pop() from %s ACIA input queue\n", q == NULL ?
    // "uninitialised" : "empty"); exit(3);
    return 0;
  }
}

int q_not_empty(void) { return q_put != q_get; }

struct vecx *GLOBAL_vecx = NULL;  // for saving dat file on quit.

// void(*vecx_render) (void);

// uint8_t rom[8192];
// uint8_t cart[32768];
// uint8_t ram[1024];

/* the sound chip registers */

// uint8_t snd_select;

// size_t vector_draw_cnt;
// vector_t vectors[VECTOR_MAX_CNT];

/* update the snd chips internal registers when VIA.ora/VIA.orb changes */

static void snd_update(vecx *vecx) {
  switch (vecx->VIA.orb & 0x18) {
    case 0x00:
      /* the sound chip is disabled */
      break;
    case 0x08:
      /* the sound chip is sending data */
      break;
    case 0x10:
      /* the sound chip is recieving data */
      if (vecx->snd_select != 14)
        e8910_write(&vecx->PSG, vecx->snd_select, vecx->VIA.ora);
      break;
    case 0x18:
      /* the sound chip is latching an address */
      if ((vecx->VIA.ora & 0xf0) == 0x00)
        vecx->snd_select = vecx->VIA.ora & 0x0f;
      break;
  }
}

static uint8_t read8_port_a(void *vecxp) {
  vecx *vvecx;
  vvecx = (vecx *)vecxp;
  if ((vvecx->VIA.orb & 0x18) == 0x08) {
    /* the snd chip is driving port a */
    return e8910_read(&vvecx->PSG, vvecx->snd_select);
  } else {
    return vvecx->VIA.ora;
  }
}

static uint8_t read8_port_b(void *vecxp) {
  vecx *vvecx;
  vvecx = (vecx *)vecxp;
  /* compare signal is an input so the value does not come from
   * VIA.orb.
   */
  if (vvecx->VIA.acr & 0x80) {
    /* timer 1 has control of bit 7 */
    return (uint8_t)((vvecx->VIA.orb & 0x5f) | vvecx->VIA.t1pb7 |
                     vvecx->DAC.compare);
  } else {
    /* bit 7 is being driven by VIA.orb */
    return (uint8_t)((vvecx->VIA.orb & 0xdf) | vvecx->DAC.compare);
  }
}

static void write8_port_a(void *vecxp, uint8_t data) {
  vecx *vvecx;
  vvecx = (vecx *)vecxp;
  snd_update(vvecx);

  /* output of port a feeds directly into the dac which then
   * feeds the x axis sample and hold.
   */
  vvecx->DAC.xsh = data ^ 0x80;
  dac_update(&vvecx->DAC);
}

static void write8_port_b(void *vecxp, uint8_t data) {
  vecx *vvecx;
  vvecx = (vecx *)vecxp;
  (void)data;
  snd_update(vvecx);
  dac_update(&vvecx->DAC);
}

static uint8_t PIA_Port[2] = {0, 0};

static uint8_t read8(void *vecxp, uint16_t address) {
  vecx *vvecx;
  uint8_t data = 0xff;
  vvecx = (vecx *)vecxp;

  if ((address & 0xFF00) == 0xE000) {  // ACIA

    int rc;
    char input;

#ifdef CONSOLE_IO
    rc = read(0 /* stdin */, &input, 1);
    if (vvecx->CPU.reg_pc == 0xfdd9 && input == ('c'&31)) exit(0); // exit to Linux on ^C *only* if in S-BUG...
    if (rc > 0) q_push(input);  // transfer any pending input to internal buffer
#else
    // q_push called in main loop.
#endif

    if ((address & 1) == 0) {  // read $E004: MC6850 Status Register (R)
      /*
         0    Receive Data  (0=No data, 1=Data can be read)
         1    Transmit Data (0=Busy, 1=Ready/Empty, Data can be written)
         2    /DCD level
         3    /CTS level
         4    Receive Framing Error (1=Error)
         5    Receive Overrun Error (1=Error)
         6    Receive Parity Error  (1=Error)
         7    Interrupt Flag (see Control Bits 5-7) (IRQ pin is not connected)
       */
      // (bit 2 (DCD) should be set if data available for reading)
      if (q_not_empty())
        return 7;  // DCD asserted and Data is ready to be read.  +2 (TD) means
                   // 'always allowed to send'.
      return 2;    // 2 (TD) means 'always allowed to send'.
    } else {       // read $E005: MC6850 Data Register (R/W)
      /*
         (Data can be read when Status.Bit0=1, and written when Status.Bit1=1.)

         0-7  Data

         Rx has a 1-stage fifo, plus 1 shift register (2 stages in total)
         Tx has a 1-stage fifo, plus 1 shift register (2 stages in total)
       */
      input = q_pop();
      if (vvecx->CPU.reg_pc == 0xfddf && input == ('c'&31)) exit(0); // exit to Linux on ^C *only* if in S-BUG...
      return input;
    }
  } else if ((address & 0xFF00) == 0xE100) {  // Read PIO control and data
                                              // registers for A and B sides.
    /*
           J2

        +5v   +5v
        e     e
        cb2   cb1
        pb7   pb6
        pb5   pb4
        pb3   pb2
        pb1   pb0
        ca2   ca1
        pa7   pa6
        pa5   pa4
        pa3   pa2
        pa1   pa0
        0v    0v   <-- pin 1
     */

    switch (address & 3) {  // read 6820/6821 PIA ADDRESSES
      case 0:  // write PIADRA/PIADD(R)A ($E100) 6821 PIA DATA REGISTER A
        /*
          CLR     PIACRA          ; ADDRESS DATA DIRECTION REGISTERS
          CLR     PIADDA          ; MAKE PORT A INPUT
        */
        PIA_Port[0] = data;
        break;
      case 1:  // read PIACRA         ($E101) 6821 PIA CONTROL REGISTER A
        /*
          LDA     #00101111b
          STA     PIACRA          ; SET PORT A AS FOLLOWS:
                                  ; BITS  7,6 NOT USED
                                  ; BIT 5 = 1 TO MAKE CA2 OUTPUT
                                  ; BIT 4 = 0 TO MAKE CA2 A PULSE
                                  ; BIT 3 = 1 TO MAKE CA2 A BRIEF
          INPUT-ACKNOWLEDGE ; BIT 2 = 1 TO ADDRESS DATA REGISTER ; BIT 1 = 1 TO
          MAKE CA1 ACTIVE HIGH ; BIT 0 = 1 TO ENABLE CA1 INTERRUPTS
         */
        break;

      case 2:  // read PIADRB/PIADDRB ($E102) 6821 PIA DATA REGISTER B
        /*
          CLR     PIACRB
          LDA     #$FF            ; MAKE PORT B OUTPUT
          STA     PIADDB
        */
        break;
      case 3:  // read PIACRB         ($E103) 6821 PIA CONTROL REGISTER B
        /*
          LDA     #00100111b
          STA     PIACRB          ; SET PORT B AS FOLLOWS:
                                  ; BITS  7, 6 NOT USED
                                  ; BIT 5 = 1 TO MAKE CB2 OUTPUT
                                  ; BIT 4 = 0 TO MAKE CB2 A PULSE
                                  ; BIT 3 = 0 TO MAKE CB2 A LONG OUTPUT-BUFFER
          FULL ; BIT 2 = 1 TO ADDRESS DATA REGISTER ; BIT 1 = 1 TO MAKE CB1
          ACTIVE HIGH ; BIT 0 = 1 TO ENABLE CB1 INTERRUPTS
         */
        break;
    }
  } else {
    data = vvecx->memory[address];
  }
#ifdef NEVER
  if ((address & 0xe000) == 0xe000) {
    /* rom */
    data = vvecx->rom[address & 0x1fff];
  } else if ((address & 0xe000) == 0xc000) {
    if (address & 0x800) {
      /* ram */
      data = vvecx->ram[address & 0x3ff];
    } else if (address & 0x1000) {
      /* io */
      data = via_read(
          &vvecx->VIA,
          address);  // How similar is the Vectrex VIA to the Motorola PIO?
    }
  } else if (address < 0x8000) {
    /* cartridge */
    data = vvecx->cart[address];
  }
#endif

  return data;
}

static void write8(void *vecxp, uint16_t address, uint8_t data) {
  vecx *vvecx;
  vvecx = (vecx *)vecxp;

  if ((address & 0xFF00) == 0xE000) {  // ACIA
    if ((address & 1) == 0) {          // MC6850 Control Register (W)
      char *Parity[8] = {"7e2", "7o2", "7e1", "7o1", "8n2", "8n1", "8e1", "8o1"};
      char *Flows[4] = { "Output RTS active low, Tx Interrupts disabled",
                         "Output RTS active low, Tx Interrupts enabled"
                         "Output RTS active high, Tx Interrupts disabled"
                         "Output RTS active low, Tx Interrupts disabled, Break sent" };
      char *Ints[2] = {  "Rx Interrupts Disabled",
                         "Rx Interrupts Enabled on buffer full/buffer overrun" };
      int Rate, Mode, Flow, Int;
      Rate = data & 3;
      Mode = (data >> 2) & 7;
      Flow = (data >> 5) & 3;
      Int = (data >> 7) & 1;
      /*
         0-1  Baudrate (0..3)
              0 = CLK/64
              1 = CLK/16
              2 = CLK/1
              3 = RESET
         2-4  Mode (data/stop bits, parity) (0..7)
              0 = 7e2
              1 = 7o2
              2 = 7e1
              3 = 7o1
              4 = 8n2
              5 = 8n1
              6 = 8e1
              7 = 8o1
         5-6  Transmit Interrupt/RTS/Break control (0..3)
              0 = Output /RTS=low,  and disable Tx Interrupt
              1 = Output /RTS=low,  and enable Tx Interrupt
              2 = Output /RTS=high, and disable Tx Interrupt
              3 = Output /RTS=low,  and disable Tx Interrupt, and send a Break
         7    Receive Interrupt (0..1)
              0 = Disable
              1 = Enable on buffer full/buffer overrun
       */
      static int first = 0;
      if ((data != 3) && (serial_port != 4)) fprintf(stderr, "Serial Port opened: %d Baud, %s; %s; %s\n",
                             VECTREX_MHZ / (Rate==0?64: Rate==1?16 : 1),
                             Parity[Mode], Flows[Flow], Ints[Int]); //$e004 ACIA control register
      first = 1;
    } else {  // MC6850 Data Register (R/W)
      /*
         (Data can be read when Status.Bit0=1, and written when Status.Bit1=1.)

         0-7  Data
      */
      // $e005 ACIA transmit register
      // fprintf(stderr, "ACIA: Send %d\n", data);
#ifdef CONSOLE_IO
      if (data) fputc(data, stderr);
#else
      if (data) {
	if (serial_port != 4) fputc(data, stderr); // copy to console for diagnostic purposes. Unless console is tty.
	int written = write(Data_To_USB_Fd, &data, 1); // unfortunately this may also be in O_NONBLOCK mode as it is tied to the input stream!
	if (written == 0) {
	  fprintf(stderr, "[FAILED]");
	}
      }
#endif
    }
  } else if ((address & 0xFF00) == 0xE100) {  // PIO
    // control and data registers for A and B sides.

    switch (address & 3) {  // 6820/6821 PIA ADDRESSES
      case 0:               // PIADRA/PIADD(R)A ($E100) 6821 PIA DATA REGISTER A
        /*
          CLR     PIACRA          ; ADDRESS DATA DIRECTION REGISTERS
          CLR     PIADDA          ; MAKE PORT A INPUT
        */
        PIA_Port[0] = data;
        break;
      case 1:  // PIACRA         ($E101) 6821 PIA CONTROL REGISTER A
        /*
          LDA     #00101111b
          STA     PIACRA          ; SET PORT A AS FOLLOWS:
                                  ; BITS  7,6 NOT USED
                                  ; BIT 5 = 1 TO MAKE CA2 OUTPUT
                                  ; BIT 4 = 0 TO MAKE CA2 A PULSE
                                  ; BIT 3 = 1 TO MAKE CA2 A BRIEF
          INPUT-ACKNOWLEDGE ; BIT 2 = 1 TO ADDRESS DATA REGISTER ; BIT 1 = 1 TO
          MAKE CA1 ACTIVE HIGH ; BIT 0 = 1 TO ENABLE CA1 INTERRUPTS
         */
        break;

      case 2:  // PIADRB/PIADDRB ($E102) 6821 PIA DATA REGISTER B
        /*
          CLR     PIACRB
          LDA     #$FF            ; MAKE PORT B OUTPUT
          STA     PIADDB
        */

        // It looks like the first write to this register after writing (BIT 2 =
        // 1 TO ADDRESS DATA REGISTER) to PIACRA treats it as PIADDRB (data
        // directon register B), but subsequent writes treat it as PIADRB (data
        // register B)
        PIA_Port[1] = data;
        break;
      case 3:  // PIACRB         ($E103) 6821 PIA CONTROL REGISTER B
        /*
          LDA     #00100111b
          STA     PIACRB          ; SET PORT B AS FOLLOWS:
                                  ; BITS  7, 6 NOT USED
                                  ; BIT 5 = 1 TO MAKE CB2 OUTPUT
                                  ; BIT 4 = 0 TO MAKE CB2 A PULSE
                                  ; BIT 3 = 0 TO MAKE CB2 A LONG OUTPUT-BUFFER
          FULL ; BIT 2 = 1 TO ADDRESS DATA REGISTER ; BIT 1 = 1 TO MAKE CB1
          ACTIVE HIGH ; BIT 0 = 1 TO ENABLE CB1 INTERRUPTS
         */
        break;
    }
  } else {
    vvecx->memory[address] = data;
  }

#ifdef NEVER
  if ((address & 0xe000) == 0xe000) {
    /* rom */
  } else if ((address & 0xe000) == 0xc000) {
    /* it is possible for both ram and io to be written at the same! */

    if (address & 0x800) {
      vvecx->ram[address & 0x3ff] = data;
    }

    if (address & 0x1000) {
      via_write(&vvecx->VIA, address, data);
    }
  } else if (address < 0x8000) {
    /* cartridge */
  }
#endif
}

void addline(void *vecxp, int32_t x0, int32_t y0, int32_t x1, int32_t y1,
             uint8_t color) {
  vecx *vvecx;
  vvecx = (vecx *)vecxp;
  vvecx->vectors[vvecx->vector_draw_cnt].x0 = x0;
  vvecx->vectors[vvecx->vector_draw_cnt].y0 = y0;
  vvecx->vectors[vvecx->vector_draw_cnt].x1 = x1;
  vvecx->vectors[vvecx->vector_draw_cnt].y1 = y1;
  vvecx->vectors[vvecx->vector_draw_cnt].color = color;
  vvecx->vector_draw_cnt++;
}

void vecx_input(vecx *vecx, uint8_t key, uint8_t value) {
  uint8_t psg_io = e8910_read(&vecx->PSG, 14);
  switch (key) {
    case VECTREX_PAD1_BUTTON1:
      e8910_write(&vecx->PSG, 14, value ? psg_io & ~0x01 : psg_io | 0x01);
      break;
    case VECTREX_PAD1_BUTTON2:
      e8910_write(&vecx->PSG, 14, value ? psg_io & ~0x02 : psg_io | 0x02);
      break;
    case VECTREX_PAD1_BUTTON3:
      e8910_write(&vecx->PSG, 14, value ? psg_io & ~0x04 : psg_io | 0x04);
      break;
    case VECTREX_PAD1_BUTTON4:
      e8910_write(&vecx->PSG, 14, value ? psg_io & ~0x08 : psg_io | 0x08);
      break;

    case VECTREX_PAD2_BUTTON1:
      e8910_write(&vecx->PSG, 14, value ? psg_io & ~0x10 : psg_io | 0x10);
      break;
    case VECTREX_PAD2_BUTTON2:
      e8910_write(&vecx->PSG, 14, value ? psg_io & ~0x20 : psg_io | 0x20);
      break;
    case VECTREX_PAD2_BUTTON3:
      e8910_write(&vecx->PSG, 14, value ? psg_io & ~0x40 : psg_io | 0x40);
      break;
    case VECTREX_PAD2_BUTTON4:
      e8910_write(&vecx->PSG, 14, value ? psg_io & ~0x80 : psg_io | 0x80);
      break;

    case VECTREX_PAD1_X:
      vecx->DAC.jch0 = value;
      break;
    case VECTREX_PAD1_Y:
      vecx->DAC.jch1 = value;
      break;
    case VECTREX_PAD2_X:
      vecx->DAC.jch2 = value;
      break;
    case VECTREX_PAD2_Y:
      vecx->DAC.jch3 = value;
      break;
  }
}

/* This is where the main vecx data structure in initialised.  Add
 * initialisation for our extra fields here. */
void vecx_reset(vecx *vecx) {
  {
    int addr;
    GLOBAL_vecx = vecx;  // because it is not passed to anywhere useful when we
                         // intercept 'Quit'
    for (addr = 0; addr <= 0xFFFF; addr++) {
      vecx->CPU.F[addr] =
          0;  // mark the ways in which ram/rom addresses are used.
      vecx->CPU.DP[addr] =
          0x100;  // Note the value of DP when executing code at this address.
                  // 100 is unassigned 200 is more than one value seen.
      vecx->CPU.IndexBase[addr] =
          0;  // When loading table,x for any address, save the value of 'table'
              // for this address.
      vecx->CPU.Index[addr] = 0;  // When loading table,x for any address, save
                                  // the value of 'x' for this address.
      vecx->CPU.disasm[addr] = NULL;  // cached disassembly
    }
    vecx->CPU.memory = vecx->memory;  // sorry, badly structured. Couldn't
                                      // always get to memory from CPU...
  }

  /* ram */
  // for (int r = 0; r < 1024; r++) vecx->ram[r] = (uint8_t)rand();

  e8910_reset(&vecx->PSG);

  vecx->snd_select = 0;

  vecx->DAC.add_line = addline;
  vecx->DAC.userdata = (void *)vecx;
  vecx->DAC.VIA = &vecx->VIA;

  dac_reset(&vecx->DAC);

  vecx->vector_draw_cnt = 0;
  vecx->fcycles = FCYCLES_INIT;

  vecx->VIA.read8_port_a = read8_port_a;
  vecx->VIA.read8_port_b = read8_port_b;
  vecx->VIA.write8_port_a = write8_port_a;
  vecx->VIA.write8_port_b = write8_port_b;
  vecx->VIA.userdata = (void *)vecx;

  via_reset(&vecx->VIA);

  vecx->CPU.read8 = read8;
  vecx->CPU.write8 = write8;
  vecx->CPU.userdata = (void *)vecx;

  e6809_reset(&vecx->CPU);
}

void vecx_emu(vecx *vecx, int32_t cycles) {
  while (cycles > 0) {
    uint16_t icycles = e6809_sstep(&vecx->CPU, vecx->VIA.ifr & 0x80, 0);

    for (uint16_t c = 0; c < icycles; c++) {
      via_sstep0(&vecx->VIA);
      dac_sstep(&vecx->DAC);
      via_sstep1(&vecx->VIA);
    }

    cycles -= (int32_t)icycles;

    vecx->fcycles -= (int32_t)icycles;

    if (vecx->fcycles < 0) {
      vecx->fcycles += FCYCLES_INIT;
      vecx->render();

      /* everything that was drawn during this pass
       * now is being removed.
       */
      vecx->vector_draw_cnt = 0;
    }
  }
}
